/**
= Creative Commons Lizenzvertrag =
Diese Software ist von der archium GmbH, Gera ist lizenziert unter einer Creative Commons Namensnennung - Nicht kommerziell - Keine Bearbeitungen 4.0 International Lizenz. (http://creativecommons.org/licenses/by-nc-nd/4.0/deed.de)
Individuelle über diese Lizenz hinausgehende Berechtigungen können Sie unter https://archium.org erhalten.

= Creative Commons License =
Software by archium GmbH, Gera is licensed under a Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International License. (http://creativecommons.org/licenses/by-nc-nd/4.0/)
Individual permissions beyond the scope of this license may be available at https://archium.org.
**/

package svg

import (
	"fmt"
	"log"
	"reflect"
	"strconv"
	"strings"

	//
	def "Toolbox/defaults"
	tol "Toolbox/goDataverse/tools"
	pog "Toolbox/postgres"
)

type TSVGobject struct {
	Name   string
	Class  string
	Trans  string
	Width  int
	Height int
	Style  string
	Anima  string
	//
	Object string
	Total  string
}

var SVG_object TSVGobject
var SVGViewbox string = "0 0 32 32"
var SVGIcons tol.SVGMap = make(tol.SVGMap)

func SVG_gethtml_all(_viewbox string) TSVGobject {
	var _svg TSVGobject = SVG_object
	if reflect.TypeOf(_svg).Name() == "TSVGobject" {
		w := strconv.Itoa(_svg.Width)
		h := strconv.Itoa(_svg.Height)
		svgs := `{`
		var icount int64 = 0
		var namelist string = ""
		for key, value := range SVGIcons {
			namelist = namelist + key + ";"
			var svg string = ""
			svg = svg + `<svg id="` + key + `" width="` + w + `" height="` + h + `" viewbox="` + _viewbox + `" style="vertical-align:text-bottom;">`
			svg = svg + `<g transform="` + _svg.Trans + `">`
			svg = svg + `<path class="` + _svg.Class + `" style="` + _svg.Style + `" d="` + value + `" />`
			svg = svg + `</g>`
			svg = svg + _svg.Anima
			svg = svg + `</svg>`
			//
			if icount > 0 {
				svgs = svgs + `,`
			}
			svgs = svgs + `"` + key + `":"` + strings.ReplaceAll(svg, `"`, `\"`) + `"`
			icount = icount + 1
		}
		svgs = svgs + `}`
		_svg.Object = svgs
		_svg.Total = fmt.Sprintf("%v", icount)
		// log.Println("SVG_object:", len(_svg.Object), "\n")
		log.Println("SVG_keys:", namelist)
	}
	//
	return _svg
}

func SVG_gethtml_single(_name string, _width, _height int, _matrix, _viewbox string) string {
	_svg := SVG_object
	var svg string = ""
	if reflect.TypeOf(_svg).Name() == "TSVGobject" {
		w := strconv.Itoa(_width)
		h := strconv.Itoa(_height)
		value := SVGIcons[_name]
		if reflect.TypeOf(value).Name() == "string" {
			svg = svg + "<svg id=\"" + _name + "\" width=\"" + w + "\" height=\"" + h + "\" viewbox=\"" + _viewbox + "\" style=\"vertical-align:text-bottom;\">"
			svg = svg + "<g transform=\"" + _matrix + "\">"
			svg = svg + "<path class=\"" + _svg.Class + "\" style=\"" + _svg.Style + "\" d=\"" + value + "\" />"
			svg = svg + "</g>"
			svg = svg + _svg.Anima
			svg = svg + "</svg>"
			//
			log.Println("SVG("+_name+"):", svg)
		}
	}
	//
	return svg
}

func GetSVG_ByParams(_fvp tol.TFVParams) (string, error) {
	var svgs string = `{"total":%total%, "data":[%svgs%]}`
	var str string = ""
	var icount int = 0
	conn, err := pog.GetConnection()
	if err == nil {
		defer pog.CloseConnection(conn)
		err = pog.PingConnection(conn)
		if err == nil {
			var svgsql string = def.DEF_sqlgetsvg
			svgsql = strings.ReplaceAll(svgsql, "%shadow%", "true")
			svgsql = strings.ReplaceAll(svgsql, "%width%", _fvp.FP_widt)
			svgsql = strings.ReplaceAll(svgsql, "%height%", _fvp.FP_heit)
			svgsql = strings.ReplaceAll(svgsql, "%matrix%", _fvp.FP_matx)
			svgsql = strings.ReplaceAll(svgsql, "%viewbox%", _fvp.FP_view)
			svgsql = strings.ReplaceAll(svgsql, "%names%", _fvp.FP_who)
			// log.Println(" <<< GetSVG_ByParams:", svgsql)
			svgrows, err := conn.Query(svgsql)
			if err != nil {
				def.LogError("GetSVG_ByParams()", err)
				return svgs, err
			}
			defer svgrows.Close()
			for svgrows.Next() {
				var s string = ""
				err := svgrows.Scan(&s)
				if err != nil {
					svgs = strings.ReplaceAll(svgs, "%total%", fmt.Sprintf("%v", icount))
					svgs = strings.ReplaceAll(svgs, "%svgs%", fmt.Sprintf("%v", str))
					//
					return svgs, err
				}
				if icount > 0 {
					str = str + ",\n"
				}
				str = str + fmt.Sprintf("%v", s)
				icount++
			}
			svgs = strings.ReplaceAll(svgs, "%total%", fmt.Sprintf("%v", icount))
			svgs = strings.ReplaceAll(svgs, "%svgs%", fmt.Sprintf("%v", str))
			//
			return svgs, err
		}
	}
	svgs = strings.ReplaceAll(svgs, "%total%", fmt.Sprintf("%v", icount))
	svgs = strings.ReplaceAll(svgs, "%svgs%", fmt.Sprintf("%v", str))
	//
	return svgs, err
}

func SVG_getsvg_single(_fvp tol.TFVParams) (string, error) {
	var svgs string = ""
	conn, err := pog.GetConnection()
	if err == nil {
		defer pog.CloseConnection(conn)
		err = pog.PingConnection(conn)
		if err == nil {
			var svgsql string = def.DEF_sqlgetsvg
			svgsql = strings.ReplaceAll(svgsql, "%shadow%", _fvp.FP_shad)
			svgsql = strings.ReplaceAll(svgsql, "%width%", _fvp.FP_widt)
			svgsql = strings.ReplaceAll(svgsql, "%height%", _fvp.FP_heit)
			svgsql = strings.ReplaceAll(svgsql, "%matrix%", _fvp.FP_matx)
			svgsql = strings.ReplaceAll(svgsql, "%viewbox%", _fvp.FP_view)
			svgsql = strings.ReplaceAll(svgsql, "%names%", _fvp.FP_who)
			// log.Println(" <<< SVG_getsvg_single:", svgsql)
			svgstr, err := conn.Query(svgsql)
			if err != nil {
				def.LogError("SVG_getsvg_single()", err)
				return svgs, err
			}
			defer svgstr.Close()
			for svgstr.Next() {
				if err := svgstr.Scan(&svgs); err == nil {
					return svgs, err
				}
			}
		}
	}
	//
	return svgs, err
}

func init() {
	log.Println("Initialize package - SVG..")
	//
	SVG_object.Width = 30
	SVG_object.Height = 30
	SVG_object.Trans = "matrix(1,0,0,1,0,0)"
	SVG_object.Class = "app-tb-btn-svg"
	SVG_object.Anima = ""
	//
	SVGIcons["expand"] = "M25.545,23.328,17.918,15.623,25.534,8.007,27.391,9.864,29.649,1.436,21.222,3.694,23.058,5.53,15.455,13.134,7.942,5.543,9.809,3.696,1.393,1.394,3.608,9.833,5.456,8.005,12.98,15.608,5.465,23.123,3.609,21.268,1.351,29.695,9.779,27.438,7.941,25.6,15.443,18.098,23.057,25.791,21.19,27.638,29.606,29.939,27.393,21.5z"
	SVGIcons["excel"] = "m 1.1015873,1.0754623 29.1013657,0 0,28.8740127 -29.1013657,0 z m 16.0853247,3.423355 c -4.528716,0.72951 -7.9678467,1.341398 -12.7886857,2.174829 l 0,17.4608507 12.6750087,2.15564 z M 11.915883,19.049367 10.825402,16.421886 c -0.293425,0.853816 -1.0375117,2.472607 -1.0477737,2.496381 l -1.400699,-0.114834 1.6308287,-3.471425 -1.6078387,-3.798243 1.440576,-0.139368 c 0.2185757,0.593963 0.9467607,2.372657 1.0220807,2.462394 0.07532,0.08974 0.983345,-1.91055 1.272699,-2.656773 l 1.659009,-0.241979 -2.024148,4.295881 c 0.600002,1.309347 1.291587,2.586904 1.835944,3.918055 z m 15.217785,-11.9490127 -9.207854,-0.123092 0,1.373542 3.182962,0 0,2.0461887 -3.159155,0.03484 0,0.909418 3.159155,-0.03484 0,2.04619 -3.133027,0 c -0.0046,0.330941 -0.01704,0.708518 0,0.909418 l 3.133027,0 0,2.04619 -3.124318,0 0,0.909418 3.124318,0 0,2.04619 -3.150446,0 0,0.909417 3.150446,0 0,2.04619 -3.182962,0 0,1.600896 9.207854,-0.123092 z m -5.115475,13.0728797 3.637671,0 0,2.04619 -3.637671,0 z m 0,-2.955607 3.637671,0 0,2.04619 -3.637671,0 z m 0,-2.955608 3.637671,0 0,2.04619 -3.637671,0 z m 0,-2.955608 3.637671,0 0,2.04619 -3.637671,0 z m 0,-2.9556067 3.637671,0 0,2.0461887 -3.637671,0 z"
	SVGIcons["order_read"] = "m 8.9643183,21.695662 -6.9341204,0 0,7.122455 6.9341204,0 z m 9.5186557,0 -6.934121,0 0,7.122455 6.934121,0 z m 9.518657,0 -6.934121,0 0,7.122455 6.934121,0 z m -9.833844,-17.6118863 0,3.8849741 -6.303746,0 0,-3.8849741 z m 1.891123,-1.9424879 -10.0859922,0 0,7.7699506 10.0859922,0 z M 21,16 15.268062,11.385612 9,16 l 4,0 0,4 4,0 0,-4 z"
	SVGIcons["barcodescanner"] = "M 10.638037,3.5437921 C 7.5301775,3.5477421 5.1899297,3.8687339 4.1731936,4.4305109 l -0.1796875,0.099609 0.033203,0.6035156 c 0.071603,1.3070943 0.3343567,2.3042653 0.9042968,3.4238281 0.5699981,1.1196765 1.4142675,2.105607 2.3359375,2.724609 l 0.21875,0.146485 1.5585938,0 1.5585936,0 0.326172,0.238281 c 0.464688,0.338805 1.118963,1.025822 1.34375,1.410156 0.105294,0.180035 0.216817,0.455198 0.248046,0.613281 0.138862,0.702968 0.273396,0.934118 0.654297,1.115234 0.232196,0.110443 0.798909,0.110039 0.966797,0 0.105622,-0.06921 0.143842,-0.03767 0.75586,0.607422 0.353958,0.373083 0.703832,0.776822 0.779296,0.896485 0.208927,0.331285 0.331668,0.874796 0.419922,1.867187 0.116889,1.314394 0.284562,1.919519 0.6875,2.472656 0.07224,0.09916 0.440319,0.493177 0.816406,0.876953 0.389653,0.397619 0.716675,0.780861 0.761719,0.888672 0.176094,0.421452 0.07805,0.931783 -0.279297,1.451172 -0.349,0.507255 -0.473848,1.067404 -0.40625,1.81836 0.084,0.933132 0.525002,1.460211 1.291016,1.548828 0.542716,0.06278 1.048645,-0.07697 2.003906,-0.5625 1.263383,-0.642144 2.589584,-1.709123 2.720703,-2.185547 0.115937,-0.421276 0.06721,-1.503786 -0.125,-2.824219 -0.642403,-4.413083 -0.922252,-5.491573 -1.886718,-7.271484 -0.96065,-1.772864 -0.97085,-1.797394 -0.970703,-2.150391 1.51e-4,-0.373794 0.195405,-1.157469 0.380859,-1.523437 0.09432,-0.186126 0.22003,-0.318532 0.478515,-0.501953 1.120154,-0.7948383 1.716057,-1.7327299 1.720704,-2.7050781 0.0022,-0.4851071 -0.245779,-0.8651372 -0.871094,-1.3261719 -1.720123,-1.2682223 -5.409728,-2.2763754 -9.462891,-2.5859375 -0.396661,-0.030293 -1.44004,-0.05389 -2.318359,-0.052734 z m -0.09961,0.8496094 c 1.225501,-0.00529 2.450382,0.017915 2.958985,0.072266 4.188561,0.4475892 7.630448,1.4980344 8.777344,2.6796875 0.197242,0.2032173 0.201193,0.2119369 0.173828,0.5136719 -0.112317,1.2386522 -1.735243,2.4501995 -3.740235,2.7929685 -0.552561,0.09447 -0.873403,0.09961 -5.826172,0.09961 l -5.2421865,0 -0.3535156,-0.294922 C 6.2540109,9.3895458 5.4706943,8.1241997 5.1048343,6.7312921 4.9488327,6.1373634 4.8084913,5.2009043 4.8567874,5.0750421 4.9176221,4.9165074 6.1358771,4.6553905 7.5716311,4.4910577 c 0.513547,-0.058778 1.741295,-0.092345 2.9667959,-0.097656 z m 2.576172,7.0351565 2.378907,0 2.380859,0 -0.224609,0.222656 c -0.341509,0.341509 -0.454474,0.697733 -0.449219,1.404296 0.0106,1.429013 0.378403,2.724018 1.240234,4.375 0.694195,1.32985 2.183398,3.977487 2.517578,4.476563 0.3631,0.542265 0.744997,0.990298 1,1.171875 0.165994,0.118199 0.439056,0.135497 0.748047,0.04687 l 0.166016,-0.04883 0,0.613281 c 0,0.604853 -0.0027,0.614621 -0.152344,0.785156 -0.453011,0.51595 -1.972904,1.461618 -2.925781,1.820313 -0.441214,0.166081 -0.828881,0.166639 -1.029297,-0.002 -0.192949,-0.162356 -0.269819,-0.430803 -0.269531,-0.935547 2.8e-4,-0.52781 0.04718,-0.686419 0.357422,-1.177735 0.304584,-0.482348 0.418434,-0.836027 0.419922,-1.30664 0.0011,-0.379702 -0.102015,-0.741769 -0.300782,-1.056641 -0.07135,-0.113023 -0.414387,-0.497028 -0.763672,-0.853516 C 17.495724,20.236398 17.4194,20.148657 17.263037,19.840667 17.097968,19.51554 17.009693,19.06196 16.92124,18.096526 16.76209,16.359462 16.625723,16.030927 15.581396,14.926604 14.568627,13.85564 14.5412,13.823214 14.419287,13.354339 14.255674,12.725095 13.930501,12.196745 13.376318,11.672698 l -0.261719,-0.24414 z m -9.1367179,6.955077 0,4.431641 0,4.429687 0.4003906,0 0.4003907,0 0,-4.429687 0,-4.431641 -0.4003907,0 -0.4003906,0 z m 1.2265625,0 0,4.431641 0,4.429687 0.1875,0 0.1875,0 0,-4.429687 0,-4.431641 -0.1875,0 -0.1875,0 z m 0.8007813,0 0,4.431641 0,4.429687 0.9882812,0 0.9902344,0 0,-4.429687 0,-4.431641 -0.9902344,0 -0.9882812,0 z m 2.4023437,0 0,4.431641 0,4.429687 0.8007813,0 0.8007811,0 0,-4.429687 0,-4.431641 -0.8007811,0 -0.8007813,0 z m 2.0273434,0 0,4.431641 0,4.429687 0.400391,0 0.40039,0 0,-4.429687 0,-4.431641 -0.40039,0 -0.400391,0 z m 1.177734,0 0,4.431641 0,4.429687 0.210938,0 0.21289,0 0,-4.429687 0,-4.431641 -0.21289,0 -0.210938,0 z m 0.800781,0 0,4.431641 0,4.429687 0.800782,0 0.800781,0 0,-4.429687 0,-4.431641 -0.800781,0 -0.800782,0 z"
	SVGIcons["leather"] = "M 14.050204,28.953458 C 13.3355,27.606945 12.011089,27.147471 10.867361,27.05813 8.5361678,27.205923 7.9600004,27.817646 6.542961,28.724555 L 3.0869069,24.88374 C 4.9705589,22.855924 6.53581,20.165164 6.7710699,17.709895 6.5342922,14.179461 5.3368386,11.37425 3.3460485,8.9573698 L 6.2846312,5.8100125 c 0.818771,0.6324049 2.9003052,1.3023618 4.4219528,0.9472896 2.025292,-1.0292149 1.610114,-2.1330545 1.93754,-3.851411 l 4.891826,0 c 0.140256,1.3049506 0.334673,2.5276313 1.545412,3.688806 2.222194,0.8910695 3.174296,0.2835914 4.757416,-0.8301509 l 2.830511,3.8082905 c -1.520087,2.3485013 -3.058407,5.2608733 -3.022306,7.7898813 0.305669,3.374029 1.851699,5.410406 3.508014,7.665514 l -3.087715,3.552905 c -1.325501,-0.737453 -2.177952,-1.868903 -4.424786,-1.601714 -1.362486,-0.08382 -2.480383,0.858732 -3.080204,1.984113 z"
	SVGIcons["drum"] = "M 24.297387,3.9447261 7.9575798,4.0335348 c -0.6638766,0 -1.2802873,0.3723424 -1.814456,0.9693276 C 5.6089604,5.5998404 5.1354199,6.435398 4.7290282,7.4674707 3.9162476,9.5315976 3.3778145,12.386285 3.3392199,15.664086 c 0.1825983,2.782486 0.5797267,5.714594 1.5660507,8.058533 0.4578931,1.061727 0.9809749,1.938287 1.5403781,2.568001 0.5594059,0.629707 1.1694599,1.038022 1.8262506,1.038022 l 16.2940137,-0.02569 c 0.645592,0 1.230096,-0.409208 1.749229,-1.03316 0.519139,-0.62396 0.991382,-1.489229 1.400914,-2.538164 0.819063,-2.097866 1.3794,-4.930765 1.3794,-7.974583 0,-3.128305 -0.499621,-6.02643 -1.324586,-8.1737184 C 27.358389,6.5096708 26.866158,5.6221001 26.292248,4.9834085 25.718336,4.3447099 25.039991,3.9446935 24.299469,3.9446935 l -0.0019,0 z m 0.0019,0.7105137 c 0.485748,0 0.9754,0.2596813 1.464052,0.803495 0.488655,0.5438197 0.953578,1.3611881 1.344708,2.379259 0.782271,2.0361485 1.276712,4.8669572 1.276712,7.9190732 0,2.956124 -0.550963,5.718304 -1.330831,7.715767 -0.389933,0.998737 -0.838415,1.80638 -1.285035,2.343183 -0.446624,0.536803 -0.725849,0.776437 -1.203162,0.776437 -1.517083,0 -2.827784,-3.848569 -3.371486,-10.48429 0.215567,-6.620935 1.426187,-11.4529242 3.105042,-11.4529242 z m 0.67582,9.2700272 a 0.53381198,1.7270387 0 0 0 -0.534271,1.727031 0.53381198,1.7270387 0 0 0 0.534271,1.727024 0.53381198,1.7270387 0 0 0 0.533586,-1.727024 0.53381198,1.7270387 0 0 0 -0.533586,-1.727031 z"
	SVGIcons["drum2"] = "M 24.296875,3.9453125 7.9570312,4.0332031 c -0.6638766,0 -1.2802844,0.3717648 -1.8144531,0.96875 C 5.7554843,5.434567 5.4661037,5.917958 5.1451657,6.5851443 l 17.5275023,0 c 0.478931,-1.2710123 0.976437,-1.9308474 1.62616,-1.9308474 0.485748,0 0.976192,0.2608738 1.464844,0.8046875 0.488655,0.5438197 0.95262,1.3608353 1.34375,2.3789062 0.782271,2.0361485 1.277344,4.8678064 1.277344,7.9199214 0,2.956126 -0.550212,5.717381 -1.330078,7.714844 -0.389933,0.998737 -0.838537,1.806947 -1.285157,2.34375 -0.446624,0.536803 -0.725812,0.775391 -1.203125,0.775391 -0.682654,0 -1.523791,-1.326865 -2.071987,-2.78962 l -17.5262271,0.02232 c 0.3855037,0.785192 1.030912,1.964226 1.4771206,2.466518 0.5594059,0.629707 1.1693812,1.037109 1.8261719,1.037109 l 16.2949216,-0.02539 c 0.645592,0 1.228914,-0.409251 1.748047,-1.033203 0.519139,-0.62396 0.992812,-1.488174 1.402344,-2.537109 0.819063,-2.097866 1.378906,-4.930788 1.378906,-7.97461 0,-3.128302 -0.499254,-6.0265392 -1.324219,-8.1738276 C 27.359003,6.5103266 26.866879,5.6230666 26.292969,4.984375 25.719057,4.3456764 25.03935,3.9453125 24.298828,3.9453125 Z M 22.57179,6.8299195 5.0019531,6.859375 c 0,0 -0.1637022,0.3287381 -0.2734375,0.6074219 C 4.2368159,8.7155111 3.9359501,9.884657 3.6953125,11.630859 L 21.537109,11.65625 C 21.742517,9.7065323 22.121726,8.0677803 22.57179,6.8299195 Z M 3.6230469,12.03125 c -0.1566397,1.138593 -0.2679834,2.340211 -0.2832031,3.632812 0.037724,0.57486 0.1377527,1.558939 0.1969866,2.140068 L 21.301898,17.75949 c -0.04444,-0.429328 -0.07118,-1.19607 -0.108539,-1.652065 0.04967,-1.525628 0.148131,-2.850858 0.275391,-4.076172 z m 21.3515621,1.894531 c -0.294521,0.0011 -0.533122,0.7737 -0.533203,1.726563 8.1e-5,0.952862 0.238682,1.725475 0.533203,1.726562 0.294521,-0.0011 0.533122,-0.7737 0.533203,-1.726562 -8.1e-5,-0.952863 -0.238682,-1.725476 -0.533203,-1.726563 z M 3.6132812,18.402344 c 0.2246978,1.741242 0.5791478,3.441012 1.1523438,4.935547 l 17.599609,0 c -0.376567,-1.29497 -0.698028,-2.950438 -0.939453,-4.935547 z"
	SVGIcons["help"] = "m 15.603663,28.164863 c -1.577387,0 -2.856728,-1.279341 -2.856728,-2.857299 0,-1.57713 1.279341,-2.856092 2.856728,-2.856092 1.577765,0 2.856663,1.278898 2.856663,2.856092 0,1.577958 -1.278898,2.857299 -2.856663,2.857299 z m 2.245985,-8.443935 0,0.513868 c 0,0 -4.28306,0 -4.709127,0 l 0,-0.513868 c 0,-1.454181 0.212145,-3.319586 1.900425,-4.941382 1.688342,-1.621984 3.798183,-2.961847 3.798183,-4.9880106 0,-2.2400216 -1.554481,-3.424837 -3.514541,-3.424837 -3.266109,0 -3.47946,3.3872804 -3.559965,4.1330876 l -4.5802798,0 C 7.3063389,6.9677341 8.7992857,2.0273662 15.351931,2.0273662 c 5.678753,0 8.23179,3.8031965 8.23179,7.3696326 0,5.6766602 -5.734073,6.6627152 -5.734073,10.3239292 z"
	SVGIcons["disk"] = "m 18.933211,7.3245112 2.997,0 0,4.9999998 -2.997,0 0,-4.9999998 z m 8.997,1 0,19.9999978 -24,0 0,-23.9999978 20,0 4,4 z m -19,4.9999998 14,0 0,-6.9999998 -14,0 0,6.9999998 z m 16,4 -18,0 0,8.999998 18,0 0,-8.999998 z"
	SVGIcons["disk2"] = "m 18.934211,7.3245101 2.996,0 0,4.9999999 -2.996,0 0,-4.9999999 z m 8.996,1 0,19.9999999 -24,0 0,-23.9999999 20,0 4,4 z m -19,4.9999999 14,0 0,-6.9999999 -14,0 0,6.9999999 z m 16,4 -18,0 0,9 18,0 0,-9 z m -2,2 -14,0 0,1 14,0 0,-1 z m 0,2 -14,0 0,1 14,0 0,-1 z m 0,2 -14,0 0,1 14,0 0,-1 z"
	SVGIcons["disk3"] = "m 16.930211,7.3245101 2.996,0 0,4.9999999 -2.996,0 0,-4.9999999 z m 11,1 0,19.9999999 -24,0 0,-23.9999999 20,0 4,4 z m -17,4.9999999 10,0 0,-6.9999999 -10,0 0,6.9999999 z m 15,-4.1709999 -2.828,-2.829 -0.172,0 0,8.9999999 -14,0 0,-8.9999999 -3,0 0,19.9999999 20,0 0,-17.1709999 z"
	SVGIcons["disk4"] = "m 16.930211,7.3245101 2.996,0 0,4.9999999 -2.996,0 0,-4.9999999 z m 11,1 0,19.9999999 -24,0 0,-23.9999999 20,0 4,4 z m -17,4.9999999 10,0 0,-6.9999999 -10,0 0,6.9999999 z m 15,-4.1709999 -2.828,-2.829 -0.172,0 0,8.9999999 -14,0 0,-8.9999999 -3,0 0,19.9999999 20,0 0,-17.1709999 z m -3,10.1709999 -14,0 0,1 14,0 0,-1 z m 0,2 -14,0 0,1 14,0 0,-1 z m 0,2 -14,0 0,1 14,0 0,-1 z"

	SVGIcons["flask"] = "m 10.243881,2.9246691 c 0,-0.72767 0.589855,-1.317526 1.317524,-1.317526 l 8.069851,0 c 0.727671,0 1.317527,0.589856 1.317527,1.317526 0,0.7276698 -0.589856,1.317526 -1.317527,1.317526 l -8.069851,0 c -0.727669,0 -1.317524,-0.5898562 -1.317524,-1.317526 z M 23.103465,28.748181 c 2.033339,0 3.371282,-2.128068 2.490255,-3.960549 L 21.321837,15.903224 C 20.133169,13.431019 19.51597,10.723041 19.51597,7.9799508 l 0,-1.7614664 -2.620227,0 c 0,1.8666054 -0.315217,5.9045606 2.051389,10.8266406 2.374046,4.937428 3.266214,6.829132 3.74566,7.825116 0.533925,1.10929 -0.989264,1.691309 -1.504681,0.613374 l -4.669376,-9.800813 -3.674978,0 c 1.671742,-4.199483 1.449412,-7.5948802 1.449412,-9.4642515 l -2.616476,0 0,1.7614663 c 0,2.7430902 -0.617194,5.4510022 -1.8058688,7.9231412 L 5.59901,24.787697 c -0.8811604,1.832416 0.457052,3.960484 2.490256,3.960484 l 15.014199,0 z"
	SVGIcons["flask2"] = "m 11.018211,2.6807418 c 0,-0.6260568 0.508105,-1.1341608 1.134161,-1.1341608 l 6.804966,0 c 0.627191,0 1.134161,0.508104 1.134161,1.1341608 0,0.6260569 -0.50697,1.1341609 -1.134161,1.1341609 l -6.804966,0 c -0.626056,0 -1.134161,-0.508104 -1.134161,-1.1341609 z M 25.762303,25.992284 c 0,1.456263 -1.160247,2.774158 -2.749206,2.774158 l -14.917619,0 c -1.587825,0 -2.748072,-1.317895 -2.748072,-2.773023 0,-0.396957 0.0862,-0.80412 0.274467,-1.198808 l 4.245164,-8.909968 c 1.179528,-2.479276 2.285335,-5.195591 2.285335,-7.9470657 l 0,-1.8543528 2.268322,0 -0.0068,2.2683216 c -0.09868,2.9749039 -1.23624,5.7751469 -2.237704,7.9391259 l 4.17598,0 4.964222,10.207448 1.696705,0 c 0.350456,0 0.589764,-0.387883 0.427579,-0.730399 l -4.245165,-8.9077 C 18.009179,14.36827 16.79109,11.426256 16.689016,8.3515461 l 0,-2.2683216 2.268322,0 0,1.8543528 c 0,2.7514747 1.105806,5.4666557 2.286468,7.9459307 l 4.245164,8.909968 c 0.187137,0.394688 0.273333,0.801852 0.273333,1.198808 z M 12.152372,24.796879 c 0,-0.940219 -0.761022,-1.701241 -1.701241,-1.701241 -0.94022,0 -1.701242,0.761022 -1.701242,1.701241 0,0.940219 0.761022,1.701241 1.701242,1.701241 0.940219,0 1.701241,-0.761022 1.701241,-1.701241 z m 2.268322,-3.969563 c 0,-0.627191 -0.508104,-1.134161 -1.134161,-1.134161 -0.627191,0 -1.134161,0.50697 -1.134161,1.134161 0,0.627191 0.50697,1.134161 1.134161,1.134161 0.626057,0 1.134161,-0.50697 1.134161,-1.134161 z m 3.402483,3.402482 c 0,-0.626056 -0.508104,-1.13416 -1.134161,-1.13416 -0.626057,0 -1.134161,0.508104 -1.134161,1.13416 0,0.626057 0.508104,1.134161 1.134161,1.134161 0.626057,0 1.134161,-0.508104 1.134161,-1.134161 z"
	SVGIcons["flask3"] = "m 20.077615,2.7412457 c 0,0.6246047 -0.505794,1.1315303 -1.131531,1.1315303 l -6.789181,0 c -0.625736,0 -1.131534,-0.5069256 -1.131534,-1.1315303 0,-0.6246047 0.505798,-1.1315303 1.131534,-1.1315303 l 6.789181,0 c 0.625737,0 1.131531,0.5069256 1.131531,1.1315303 z M 25.462567,24.803823 21.22725,15.914522 C 20.049326,13.440997 18.946084,10.730981 18.946084,7.985889 l 0,-1.8500525 -2.26306,0 0,2.263061 c 0.101838,3.0675785 1.317101,6.0027685 2.500682,8.4887395 l 4.235317,8.887039 c 0.161809,0.341723 -0.07695,0.728706 -0.426587,0.728706 l -14.883028,0 c -0.339459,0 -0.590659,-0.383588 -0.425455,-0.726443 l 4.234197,-8.890433 c 1.184712,-2.48484 2.393186,-5.421161 2.495024,-8.4876085 l 0.0068,-2.263061 -2.26306,0 0,1.8500525 c 0,2.745092 -1.103245,5.453976 -2.281177,7.927501 L 5.64042,24.802692 c -0.186703,0.393773 -0.272699,0.801123 -0.272699,1.197159 0,1.451754 1.157555,2.766592 2.741698,2.766592 l 14.883028,0 c 1.584143,0 2.74283,-1.314838 2.74283,-2.767724 0,-0.396035 -0.086,-0.802254 -0.272699,-1.194896 z M 14.419963,23.674556 c 0,-0.938038 -0.760388,-1.697295 -1.697295,-1.697295 -0.936907,0 -1.697299,0.759257 -1.697299,1.697295 0,0.938039 0.760392,1.697296 1.697299,1.697296 0.936907,0 1.697295,-0.759257 1.697295,-1.697296 z m 2.263061,-3.960355 c 0,-0.625737 -0.506926,-1.131531 -1.131531,-1.131531 -0.625735,0 -1.13153,0.505794 -1.13153,1.131531 0,0.625735 0.505795,1.131529 1.13153,1.131529 0.624605,0 1.131531,-0.505794 1.131531,-1.131529 z m 3.394591,3.394591 c 0,-0.624606 -0.506926,-1.131531 -1.131531,-1.131531 -0.625736,0 -1.13153,0.506925 -1.13153,1.131531 0,0.624604 0.505794,1.131529 1.13153,1.131529 0.624605,0 1.131531,-0.506925 1.131531,-1.131529 z"
	SVGIcons["flask4"] = "m 20.1118,2.6807417 c 0,0.6260569 -0.508105,1.1341609 -1.134161,1.1341609 l -6.804966,0 c -0.626056,0 -1.134161,-0.508104 -1.134161,-1.1341609 0,-0.6260568 0.508105,-1.1341608 1.134161,-1.1341608 l 6.804966,0 c 0.626056,0 1.134161,0.508104 1.134161,1.1341608 z M 25.509271,24.794611 21.264107,15.884643 C 20.083445,13.405367 18.977639,10.689052 18.977639,7.9375773 l 0,-1.8543529 -2.268322,0 0,2.2683217 c 0.102074,3.0747099 1.320163,6.0167239 2.506495,8.5084749 l 4.245165,8.9077 c 0.162185,0.342516 -0.07712,0.730399 -0.426445,0.730399 l -14.918749,0 c -0.339114,0 -0.592032,-0.38448 -0.426445,-0.728131 L 11.9345,16.858887 c 1.186332,-2.490617 2.397616,-5.433765 2.49969,-8.5073409 l 0.0068,-2.2683217 -2.268322,0 0,1.8543529 c 0,2.7514747 -1.105807,5.4666557 -2.286467,7.9459317 l -4.245162,8.909967 c -0.187137,0.394688 -0.273333,0.802986 -0.273333,1.199943 0,1.455128 1.160247,2.773023 2.748072,2.773023 l 14.918749,0 c 1.587825,0 2.748072,-1.317895 2.748072,-2.774158 0,-0.396956 -0.0862,-0.80412 -0.273333,-1.197673 z"
	SVGIcons["flask5"] = "m 7.047519,26.440246 -1.826808,0 c -1.576777,0 -2.72895,-1.314838 -2.72895,-2.766591 0,-0.396036 0.0856,-0.802255 0.271431,-1.196028 L 6.978817,14.719855 C 8.151263,12.24633 9.249374,9.536314 9.249374,6.791222 l 0,-0.718521 2.252533,0 -0.0068,1.13153 C 11.393785,10.270679 10.190929,13.207 9.011733,15.69184 l -4.214499,7.758904 c -0.165562,0.342853 0.0856,0.726441 0.423477,0.726441 l 2.12527,0 c -0.396447,1.05685 -0.371669,1.572827 -0.298462,2.263061 z m 2.201855,-22.630607 6.757609,0 c 0.622827,0 1.126269,-0.506925 1.126269,-1.13153 0,-0.624604 -0.503442,-1.13153 -1.126269,-1.13153 l -6.757609,0 c -0.622825,0 -1.126268,0.506926 -1.126268,1.13153 0,0.624605 0.503443,1.13153 1.126268,1.13153 z m 13.515223,1.131532 -6.757614,0 c -0.622827,0 -1.126269,0.506925 -1.126269,1.13153 0,0.624604 0.503442,1.13153 1.126269,1.13153 l 6.757614,0 c 0.622827,0 1.126269,-0.506926 1.126269,-1.13153 0,-0.624605 -0.503442,-1.13153 -1.126269,-1.13153 z m 6.757615,20.994412 c 0,1.452886 -1.152174,2.767724 -2.730077,2.767724 l -14.813817,0 c -1.576776,0 -2.728944,-1.314838 -2.728944,-2.766592 0,-0.396035 0.0856,-0.802255 0.271429,-1.196028 2.686148,-4.222871 6.48618,-9.198209 6.48618,-14.554873 l 0,-0.718522 2.252538,0 -0.0068,1.131531 c -0.117132,3.513401 -1.680393,6.794838 -2.749222,9.052241 l 5.255171,0 4.348525,6.789182 1.684898,0 c 0.348018,0 0.58566,-0.386984 0.424604,-0.728706 -2.749223,-4.319051 -6.509835,-9.215182 -6.70468,-15.112717 l 0,-1.131531 2.252538,0 0,0.718522 c 0,5.354401 3.777507,10.298057 6.486184,14.553742 0.185834,0.393773 0.271431,0.799992 0.271431,1.196027 z"
	SVGIcons["flask6"] = "M 25.56987,24.757761 C 22.889022,20.564786 19.18826,16.793055 18.995954,11.596192 l -2.307682,0 c 0.19803,5.96609 4.020128,9.774451 6.8143,14.143707 0.16369,0.345695 -0.07784,0.737176 -0.431546,0.737176 l -1.44688,0 -5.41893,-8.012783 -4.029286,0 c 0.999309,-1.905897 2.117665,-3.907949 2.216107,-6.8681 l -2.300813,0 c -0.193452,5.200297 -3.915958,9.00179 -6.573912,13.161569 -0.188873,0.397205 -0.275869,0.809291 -0.275869,1.20993 0,1.468629 1.171011,2.798751 2.773568,2.798751 l 15.056015,0 c 1.602557,0 2.774713,-1.330122 2.774713,-2.799895 0,-0.40064 -0.087,-0.811581 -0.275869,-1.208786 z M 20.122322,7.0174591 c 0,1.2648751 -1.025636,2.2893667 -2.289367,2.2893667 -1.26373,0 -2.289366,-1.0244916 -2.289366,-2.2893667 0,-1.2648751 1.025636,-2.2893667 2.289366,-2.2893667 1.263731,0 2.289367,1.0244916 2.289367,2.2893667 z m -5.723417,-1.717025 c 0,0.9489425 -0.769227,1.717025 -1.717025,1.717025 -0.947797,0 -1.717024,-0.7680825 -1.717024,-1.717025 0,-0.9489425 0.769227,-1.717025 1.717024,-1.717025 0.947798,0 1.717025,0.7680825 1.717025,1.717025 z m 0,4.0063917 c 0,0.6318652 -0.511673,1.1446832 -1.144683,1.1446832 -0.63301,0 -1.144683,-0.512818 -1.144683,-1.1446832 0,-0.6318652 0.511673,-1.1446833 1.144683,-1.1446833 0.63301,0 1.144683,0.5128181 1.144683,1.1446833 z m 3.43405,-6.8681 c 0,0.6318652 -0.511673,1.1446833 -1.144683,1.1446833 -0.63301,0 -1.144683,-0.5128181 -1.144683,-1.1446833 0,-0.6318652 0.511673,-1.1446833 1.144683,-1.1446833 0.63301,0 1.144683,0.5128181 1.144683,1.1446833 z"
	SVGIcons["flask7"] = "m 15.136407,10.619868 -1.367798,0 c -0.0079,-1.2804678 -0.119087,-2.3828722 -0.913,-3.687157 l 1.168186,-0.7099847 c 0.976513,1.6082399 1.103539,2.9612938 1.112612,4.3971417 z m 2.205943,0 -1.363262,0 c 0.01134,-1.500495 0.141771,-2.80478 1.108076,-4.3971417 L 18.255349,6.932711 C 17.462571,8.23813 17.350289,9.3473393 17.34235,10.619868 Z M 11.133952,8.4343398 C 9.623241,7.2015071 9.5665335,5.5875962 10.497683,4.4466303 c 0.547804,-0.6725574 1.288413,-1.0116715 2.186668,-0.9980615 0.5943,-1.2430404 1.737535,-1.9019878 2.871695,-1.9019878 1.170454,0 2.286469,0.6963747 2.862623,1.9019878 0.900523,-0.01361 1.634325,0.3221017 2.186662,0.9980615 0.923207,1.1318926 0.900523,2.7401327 -0.682765,4.0251368 L 18.966468,7.7992097 c 0.440055,-0.5296529 0.860828,-1.0037322 0.860828,-1.5832884 0,-0.9753783 -1.164783,-1.8464139 -2.348847,-1.2010763 -0.0034,-1.2169547 -0.939085,-2.0346847 -1.874768,-2.0607704 -0.982183,-0.026086 -1.972306,0.8063884 -1.975708,2.0607704 -1.233967,-0.6714233 -2.35225,0.2744669 -2.35225,1.2010763 0,0.5727513 0.411701,1.0400256 0.84495,1.5628735 l -0.986721,0.655545 z M 25.486759,24.794611 C 22.830554,20.640179 19.163812,16.903119 18.973273,11.754029 l -2.286468,0 c 0.19621,5.911246 3.983173,9.6846 6.75166,14.013692 0.162185,0.342516 -0.07712,0.730399 -0.427579,0.730399 l -1.433579,0 -5.369118,-7.939126 -3.992246,0 c 0.990122,-1.888378 2.098197,-3.872025 2.195735,-6.804965 l -2.279663,0 c -0.191673,5.152493 -3.879975,8.919041 -6.513496,13.040582 -0.1871365,0.393553 -0.273333,0.801851 -0.273333,1.198808 0,1.455128 1.160247,2.773023 2.7480725,2.773023 l 14.9176275,0 c 1.587825,0 2.749206,-1.317895 2.749206,-2.774158 0,-0.396956 -0.0862,-0.80412 -0.273333,-1.197673 z M 10.448914,26.49812 c -0.939089,0 -1.701245,-0.761022 -1.701245,-1.701241 0,-0.940219 0.762156,-1.701241 1.701245,-1.701241 0.939091,0 1.701247,0.761022 1.701247,1.701241 0,0.940219 -0.762156,1.701241 -1.701247,1.701241 z m 3.969569,-3.402482 c -0.626057,0 -1.134161,-0.508105 -1.134161,-1.134161 0,-0.626057 0.508104,-1.134161 1.134161,-1.134161 0.626057,0 1.134161,0.508104 1.134161,1.134161 0,0.626056 -0.508104,1.134161 -1.134161,1.134161 z m 2.268322,3.402482 c -0.626057,0 -1.134161,-0.508104 -1.134161,-1.134161 0,-0.626056 0.508104,-1.134161 1.134161,-1.134161 0.626057,0 1.134161,0.508105 1.134161,1.134161 0,0.626057 -0.508104,1.134161 -1.134161,1.134161 z"
	SVGIcons["flask8"] = "M 25.519262,24.776186 C 22.850735,20.602483 19.166984,16.848087 18.975561,11.675111 l -2.297075,0 c 0.19712,5.938667 4.001651,9.729525 6.782979,14.078699 0.162938,0.344106 -0.07748,0.733788 -0.429561,0.733788 l -1.44023,0 -5.394024,-7.975955 -4.010766,0 c 0.994715,-1.897138 2.107931,-3.889987 2.205921,-6.836532 l -2.290238,0 c -0.192563,5.176394 -3.8979565,8.960415 -6.5436955,13.101075 -0.188004,0.39538 -0.2746,0.805571 -0.2746,1.204369 0,1.461879 1.165629,2.785887 2.76082,2.785887 l 14.9868125,0 c 1.595191,0 2.761959,-1.324008 2.761959,-2.787026 0,-0.398798 -0.0866,-0.80785 -0.274601,-1.20323 z m -15.107598,1.711412 c -0.9434365,0 -1.7091255,-0.764552 -1.7091255,-1.709133 0,-0.944581 0.765689,-1.709134 1.7091255,-1.709134 0.943442,0 1.709133,0.764553 1.709133,1.709134 0,0.944581 -0.765691,1.709133 -1.709133,1.709133 z m 3.987977,-3.418267 c -0.62896,0 -1.139421,-0.510461 -1.139421,-1.139421 0,-0.628961 0.510461,-1.139423 1.139421,-1.139423 0.628962,0 1.139423,0.510462 1.139423,1.139423 0,0.62896 -0.510461,1.139421 -1.139423,1.139421 z m 2.278845,3.418267 c -0.628961,0 -1.139422,-0.510461 -1.139422,-1.139423 0,-0.62896 0.510461,-1.139421 1.139422,-1.139421 0.628961,0 1.139422,0.510461 1.139422,1.139421 0,0.628962 -0.510461,1.139423 -1.139422,1.139423 z m 2.519262,-18.9941666 1.610004,3.0422566 -4.365127,0 0.680236,-1.3559116 1.357051,0 -0.893306,-1.686345 0.893306,-1.6874839 -1.919926,0 -1.020922,-1.6920418 -1.016364,1.7034361 -1.918788,0 0.895586,1.6760896 -0.895586,1.6760897 1.335403,0 0.680235,1.3467969 -4.349174,0 1.614561,-3.0228866 -1.614561,-3.0228866 3.443333,0 1.819658,-3.050233 1.829911,3.0308628 3.443334,0 -1.608864,3.0422568 z"
	SVGIcons["flask9"] = "M 19.303508,7.1675212 C 18.66784,6.5318534 19.12905,5.4560308 20.012368,5.4560308 c 0.256673,0 0.513346,0.098258 0.708859,0.2937707 l 4.254162,4.2531575 c 0.392029,0.392029 0.392029,1.025692 0,1.41772 -0.392028,0.392029 -1.025691,0.392029 -1.41772,0 L 19.303508,7.1675212 Z m -1.815764,9.3174448 -11.4500413,0 -0.4461706,-0.44617 C 5.3789744,15.825235 5.4621928,15.428194 5.7800267,15.315899 11.96726,13.955329 15.179688,14.051582 19.388731,9.8876574 L 17.971012,8.4699381 C 14.177057,12.273918 10.72099,12.191702 5.1102694,13.426943 c -1.0317069,0.366963 -1.6423089,1.320464 -1.6423089,2.328108 0,0.606592 0.2245893,1.220202 0.7068546,1.701464 l 9.3254669,9.325468 c 0.481263,0.481262 1.09387,0.73192 1.700461,0.73192 1.002631,0 1.960143,-0.630655 2.329111,-1.668377 1.238249,-5.631776 0.886326,-9.312433 4.695319,-13.120424 l -1.417719,-1.41772 c -1.736557,1.736556 -2.696074,3.406939 -3.31971,5.177584 z m 10.043351,1.520991 c 0,1.366585 -1.10891,2.489532 -2.476498,2.489532 -2.998868,0 -4.07068,-4.231101 0.914399,-8.021045 -0.05815,2.371221 1.562099,3.310686 1.562099,5.531513 z m -2.377238,-3.475118 c -0.34691,0.189497 -0.858251,0.699836 -0.936457,1.336507 -0.115302,0.952499 0.869281,1.233236 0.955507,-0.04412 0.04011,-0.538412 0,-0.796089 -0.01905,-1.292391 z"
	SVGIcons["filter"] = "M26.834,6.958c0-2.094-4.852-3.791-10.834-3.791c-5.983,0-10.833,1.697-10.833,3.791c0,0.429,0.213,0.84,0.588,1.224l8.662,15.002v4.899c0,0.414,0.709,0.75,1.583,0.75c0.875,0,1.584-0.336,1.584-0.75v-4.816l8.715-15.093h-0.045C26.625,7.792,26.834,7.384,26.834,6.958zM16,9.75c-6.363,0-9.833-1.845-9.833-2.792S9.637,4.167,16,4.167c6.363,0,9.834,1.844,9.834,2.791S22.363,9.75,16,9.75z"
	SVGIcons["search"] = "M29.772,26.433l-7.126-7.126c0.96-1.583,1.523-3.435,1.524-5.421C24.169,8.093,19.478,3.401,13.688,3.399C7.897,3.401,3.204,8.093,3.204,13.885c0,5.789,4.693,10.481,10.484,10.481c1.987,0,3.839-0.563,5.422-1.523l7.128,7.127L29.772,26.433zM7.203,13.885c0.006-3.582,2.903-6.478,6.484-6.486c3.579,0.008,6.478,2.904,6.484,6.486c-0.007,3.58-2.905,6.476-6.484,6.484C10.106,20.361,7.209,17.465,7.203,13.885z"

	SVGIcons["reset_sett"] = "m 6.27412,13.646205 c -0.297,0.947 -0.458,1.955 -0.458,3 0,5.52 4.481,10 10,10 5.52,0 10,-4.48 10,-10 0,-5.519 -4.48,-10 -10,-10 -2.121,0 -4.083,0.668 -5.703,1.7960001 l 1.703,2.2039999 -6.58,0 1.935,-6.012 1.718,2.223 c 1.958,-1.389 4.346,-2.211 6.927,-2.211 6.623,0 12,5.377 12,12 0,6.623 -5.377,11.988 -12,11.988 -6.623,0 -12,-5.365 -12,-11.988 0,-1.036 0.132,-2.041 0.379,-3 l 2.079,0 z m 10.35,-3.012 c 0.292,0.821 0.375,1.346 1.01,1.609 0.637,0.264 1.073,-0.052 1.854,-0.423 l 1.142,1.142 c -0.373,0.787 -0.687,1.218 -0.423,1.854 0.262,0.634 0.784,0.716 1.609,1.009 l 0,1.617 c -0.816,0.29 -1.347,0.375 -1.61,1.01 -0.264,0.636 0.052,1.071 0.424,1.853 l -1.142,1.142 c -0.79,-0.375 -1.219,-0.687 -1.85,-0.424 -0.639,0.265 -0.723,0.793 -1.014,1.611 l -1.616,0 c -0.292,-0.821 -0.375,-1.347 -1.01,-1.61 -0.637,-0.264 -1.072,0.052 -1.854,0.423 l -1.142,-1.142 c 0.366,-0.771 0.689,-1.212 0.423,-1.854 -0.263,-0.635 -0.793,-0.719 -1.609,-1.009 l 0,-1.617 c 0.817,-0.29 1.346,-0.373 1.609,-1.009 0.264,-0.637 -0.051,-1.07 -0.423,-1.854 l 1.142,-1.142 c 0.788,0.374 1.218,0.687 1.854,0.423 0.635,-0.263 0.719,-0.792 1.01,-1.609 l 1.616,0 z m -0.808,8 c -1.105,0 -2,-0.896 -2,-2 0,-1.105 0.895,-2.001 2,-2.001 1.104,0 2,0.896 2,2.001 0,1.104 -0.896,2 -2,2 z"
	SVGIcons["reset_time"] = "m 27.81612,16.634205 c 0,6.627 -5.373,12 -12,12 -6.627,0 -12,-5.373 -12,-12 l 2,0 c 0,5.514 4.486,10 10,10 5.514,0 10,-4.486 10,-10 0,-5.514 -4.486,-9.9999997 -10,-9.9999997 -2.777,0 -5.287,1.141 -7.099,2.9769995 l 2.061,2.0610002 -6.962,1.354 1.305,-7.0129997 2.179,2.18 c 2.172,-2.196 5.182,-3.559 8.516,-3.559 6.627,0 12,5.3729997 12,11.9999997 z m -13,-6 0,8 7,0 0,-2 -5,0 0,-6 -2,0 z"

	SVGIcons["sheep1"] = "m 13.405424,28.670839 c -1.35431,-0.2139 -2.5818,-1.02856 -3.23644,-2.14793 l -0.2111799,-0.36111 -0.35207,0.11442 c -0.2417,0.0786 -0.56906,0.12302 -1.0442497,0.14188 -0.89836,0.0356 -1.38029,-0.065 -2.17993,-0.45525 -0.52619,-0.2568 -0.66221,-0.35625 -1.14441,-0.83667 -0.48783,-0.48603 -0.57843,-0.60993 -0.83437,-1.14101 -0.15793,-0.32767 -0.32941,-0.78772 -0.38107,-1.02232 -0.11931,-0.54179 -0.12194,-1.39533 -0.006,-1.89299 l 0.09,-0.38586 -0.31656,-0.38194 c -0.39958,-0.48208 -0.72403,-1.14599 -0.88149,-1.80378 -0.22656,-0.94642 -0.0861,-2.13735 0.35415,-3.00324 0.36577,-0.71937 1.11713,-1.51507 1.74827,-1.85147 0.19117,-0.10189 0.34758,-0.20622 0.34758,-0.23185 0,-0.0256 -0.0382,-0.1381 -0.085,-0.24995 -0.0937,-0.22425 -0.0627,-0.21384 -0.96719,-0.32446 -0.70168,-0.0858 -1.22283,-0.19635 -1.77724,-0.377 -0.87122,-0.28386 -1.03134,-0.58917 -0.46941,-0.89502 0.38033,-0.20699 1.38292,-0.45531 2.33474,-0.5782501 l 0.77056,-0.0995 0.21005,-0.43405 c 0.11553,-0.23873 0.34377,-0.58407 0.50718,-0.76742 l 0.29714,-0.33338 0.006,-0.74486 c 0.007,-0.83388 0.0942,-1.17925 0.44823,-1.78115 0.54783,-0.93127 1.67483,-1.60832 2.6771697,-1.60832 h 0.36383 l 0.17034,-0.35546 c 0.7568499,-1.57946 2.6708999,-2.32521 4.2426699,-1.65303 0.42468,0.18163 0.48659,0.18146 0.91718,-0.003 0.19138,-0.0818 0.54328,-0.17771 0.78201,-0.21321 0.53906,-0.0802 1.14738,-0.002 1.65013,0.21321 0.43302,0.185 0.49097,0.18468 0.9363,-0.005 1.10029,-0.46916 2.3861,-0.25673 3.37001,0.55678 0.3017,0.24945 0.44119,0.32396 0.71092,0.37975 1.39422,0.28834 2.43168,1.35874 2.67263,2.75744 0.0241,0.14 0.11449,0.3116 0.23498,0.44617 0.46181,0.51577 0.78854,1.18041 0.89702,1.82468 0.0263,0.15626 0.12303,0.31864 0.32182,0.5403 0.1563,0.17429 0.37823,0.51124 0.49317,0.74877 l 0.20897,0.43189 0.76778,0.0993 c 0.94973,0.1228601 1.95221,0.3713701 2.33197,0.5780501 0.56193,0.30585 0.40181,0.61116 -0.46941,0.89502 -0.55441,0.18065 -1.07556,0.29119 -1.77724,0.37701 -0.90453,0.11061 -0.8735,0.1002 -0.96719,0.32445 -0.0467,0.11185 -0.085,0.22433 -0.085,0.24995 0,0.0256 0.15641,0.12996 0.34758,0.23185 0.63114,0.3364 1.3825,1.13211 1.74827,1.85147 0.44028,0.86589 0.58071,2.05682 0.35415,3.00324 -0.15746,0.65779 -0.48191,1.3217 -0.88149,1.80378 l -0.31656,0.38194 0.09,0.38586 c 0.11611,0.49766 0.11348,1.3512 -0.006,1.893 -0.0517,0.23459 -0.22314,0.69464 -0.38107,1.02231 -0.25594,0.53108 -0.34654,0.65498 -0.83437,1.14101 -0.4822,0.48042 -0.61822,0.57987 -1.14441,0.83668 -0.79964,0.39024 -1.28157,0.49088 -2.17993,0.45524 -0.47519,-0.0189 -0.80255,-0.0633 -1.04425,-0.14188 l -0.35207,-0.11442 -0.21118,0.36111 c -0.56559,0.9671 -1.51078,1.68231 -2.66771,2.0186 -0.36371,0.10572 -0.59189,0.13229 -1.13747,0.13243 -0.76671,2.3e-4 -1.30628,-0.11949 -1.90805,-0.42327 l -0.33755,-0.17056 -0.33677,0.17 c -0.51219,0.25856 -1.09745,0.40205 -1.73911,0.42639 -0.31281,0.0119 -0.64551,0.009 -0.73936,-0.005 z m 1.43514,-2.41174 c 0.195,-0.0888 0.45378,-0.24113 0.57509,-0.33861 0.28493,-0.22898 0.65987,-0.34258 0.95125,-0.28822 0.24082,0.0449 0.67626,0.24936 0.73558,0.34534 0.0753,0.12188 0.7419,0.4107 1.04757,0.45391 0.64111,0.0906 1.32152,-0.13182 1.81491,-0.59332 0.30972,-0.2897 0.60384,-0.79333 0.66619,-1.1407 0.1015,-0.56546 0.46697,-0.96221 1.04936,-1.1392 0.31471,-0.0956 0.69907,-0.004 1.17505,0.27883 0.2349,0.13979 0.55572,0.28503 0.71294,0.32275 0.83347,0.19994 1.83768,-0.18035 2.3147,-0.87655 0.4459,-0.65079 0.52658,-1.48872 0.20894,-2.16999 -0.20494,-0.43955 -0.25007,-0.99113 -0.10436,-1.27536 0.0552,-0.10774 0.25586,-0.32528 0.44586,-0.48342 0.4756,-0.39588 0.74288,-0.77041 0.88593,-1.24139 0.31329,-1.03141 -0.19658,-2.16205 -1.18784,-2.63404 -0.42314,-0.20148 -1.03434,-0.24743 -1.55659,-0.11702 -0.20007,0.05 -0.47893,0.088 -0.61969,0.0845 -0.14077,-0.004 -0.42229,0.0351 -0.62561,0.0858 -0.20332,0.0507 -0.46172,0.10574 -0.57423,0.12236 -0.18039,0.0266 -0.21264,0.0572 -0.27295,0.25849 -0.0376,0.12555 -0.23493,0.56101 -0.43848,0.96768 -1.4411,2.87909 -3.94992,5.05824 -5.82352,5.05824 -1.85731,0 -4.35552,-2.15583 -5.79305,-4.99912 -0.18346,-0.36289 -0.4019,-0.84161 -0.4854099,-1.06383 l -0.15185,-0.40405 -0.4674,-0.0434 c -0.25707,-0.0239 -0.5313797,-0.0371 -0.6095797,-0.0293 -0.35107,0.0347 -0.62046,0.0193 -0.91892,-0.0526 -1.3439,-0.32376 -2.64538,0.60673 -2.75994,1.97323 -0.0641,0.76404 0.26028,1.44777 0.95758,2.01867 0.44616,0.36527 0.58302,0.63945 0.54426,1.09029 -0.0174,0.20233 -0.0948,0.48442 -0.18767,0.68434 -0.31699,0.68205 -0.23615,1.51964 0.20944,2.16999 0.47702,0.6962 1.48123,1.07649 2.3147,0.87655 0.1572197,-0.0377 0.4836997,-0.18633 0.7254997,-0.33023 0.33639,-0.20018 0.5115299,-0.2676 0.7456699,-0.28701 0.70883,-0.0588 1.33675,0.43583 1.46609,1.15486 0.10568,0.58748 0.66346,1.27249 1.25765,1.54451 0.60574,0.27733 1.18875,0.28294 1.77283,0.0171 z m -0.71043,-8.91963 c 0.30718,-0.1487 0.63694,-0.46738 0.80182,-0.77488 0.17553,-0.32734 0.17567,-1.15089 2.9e-4,-1.47871 -0.41365,-0.77308 -1.29187,-1.12047 -2.08369,-0.82422 -1.18227,0.44235 -1.48173,2.0059 -0.53875,2.81305 0.5058,0.43296 1.24948,0.54112 1.82036,0.26476 z m 5.46267,0.0493 c 0.66663,-0.24942 1.08511,-0.83814 1.0876,-1.53003 0.002,-0.51491 -0.10247,-0.77319 -0.46643,-1.15478 -0.50767,-0.53224 -1.13599,-0.68771 -1.79014,-0.44296 -0.34155,0.1278 -0.72842,0.47664 -0.91473,0.82485 -0.17803,0.33274 -0.17551,1.15431 0.005,1.47871 0.43895,0.79086 1.27862,1.12372 2.07915,0.82421 z m 3.59866,-4.17974 c 0.33395,-0.16063 0.48096,-0.19907 0.76138,-0.19907 0.19112,0 0.43983,-0.0386 0.55269,-0.0857 l 0.20518,-0.0857 -0.23222,-0.16823 c -0.33344,-0.24154 -0.45055,-0.7372 -0.274,-1.15974 0.073,-0.17464 0.29591,-0.3979 0.44791,-0.44857 0.0361,-0.012 -0.0434,-0.1204001 -0.17668,-0.2408101 -0.30281,-0.27356 -0.40583,-0.51547 -0.40902,-0.9604 -0.004,-0.53962 -0.10758,-0.7533 -0.56635,-1.16687 -0.47005,-0.42373 -0.56862,-0.62112 -0.56862,-1.13867 0,-0.47889 -0.15944,-0.81809 -0.49366,-1.05025 -0.2061,-0.14318 -0.30823,-0.17002 -0.76779,-0.20177 -0.29238,-0.0202 -0.59678,-0.074 -0.67885,-0.11999 -0.0818,-0.0459 -0.294,-0.25543 -0.47147,-0.46574 -0.26442,-0.31332 -0.3719,-0.39697 -0.59526,-0.4633 -0.41495,-0.12323 -0.89965,0.003 -1.17331,0.3056 -0.42751,0.4727 -1.27873,0.47259 -1.70607,-2.2e-4 -0.18244,-0.20186 -0.53735,-0.35436 -0.82466,-0.35436 -0.28731,0 -0.64222,0.1525 -0.82466,0.35436 -0.33789,0.37383 -1.01129,0.47618 -1.42301,0.21627 -0.1254,-0.0792 -0.2643,-0.18767 -0.30866,-0.24113 -0.23046,-0.27768 -0.8413,-0.40756 -1.22575,-0.26061 -0.20854,0.0797 -0.46472,0.31551 -0.57758,0.53163 -0.0469,0.0899 -0.094,0.37141 -0.10657,0.638 -0.0233,0.49298 -0.13461,0.81294 -0.34774,1.00002 -0.35274,0.30962 -0.80421,0.35544 -1.32836,0.13482 -0.7602699,-0.31999 -1.4360096,-0.0371 -1.6511496,0.69122 -0.0807,0.27324 -0.0449,0.53142 0.11517,0.83006 0.3839597,0.71637 0.20028,1.3708 -0.48528,1.72901 -0.17066,0.0892 -0.31018,0.1749201 -0.31006,0.1905601 1.1e-4,0.0156 0.0746,0.0708 0.1656,0.12251 0.33778,0.19217 0.47754,0.63931 0.34902,1.11661 -0.0485,0.18 -0.12545,0.2862 -0.29818,0.41132 l -0.23284,0.16867 0.20518,0.0857 c 0.11286,0.0472 0.34449,0.0857 0.51475,0.0857 0.17026,0 0.4717997,0.0508 0.6700897,0.11302 0.45578,0.14287 0.59445,0.14239 0.9841199,-0.003 0.39555,-0.148 0.8016,-0.46617 1.08144,-0.84737 0.41458,-0.56479 0.81557,-0.70213 1.50831,-0.51663 0.90241,0.24167 2.27737,0.0467 2.75766,-0.39107 0.43197,-0.3937101 0.92142,-0.3833601 1.76379,0.0373 0.76841,0.38371 1.1167,0.49179 1.59245,0.49412 0.33157,0.002 0.46875,-0.0298 0.82072,-0.18793 0.88271,-0.39659 1.53986,-0.12695 1.787,0.73323 0.10479,0.36477 0.35923,0.69646 0.65353,0.85192 0.32195,0.17009 0.6434,0.14598 1.12181,-0.0841 z"
	SVGIcons["goat"] = "m 25.59315,0.5811 c -3.45756,1.11883 -6.81761,3.16115 -8.2234,6.36807 -0.37149,0.6011 -0.87255,3.18898 -1.66197,1.5909 -0.93564,-3.86612 -4.65859,-6.74709 -8.71538,-7.85984 1.91057,2.57567 3.88314,5.2564 4.52215,8.36445 -0.62435,1.42594 -2.52857,-0.30416 -3.74019,0.27419 -1.73184,0.14705 -3.37285,0.86244 -4.66927,1.90315 1.10968,2.13076 3.825,3.2564 6.33988,3.11714 1.31713,0.92209 -0.21767,2.63855 0.28447,3.90767 -0.007,2.07259 0.72818,4.08521 1.84499,5.86891 0.53326,1.44365 0.16072,3.40529 1.8158,4.31221 1.09685,1.11396 3.21341,1.96583 4.60182,0.78994 1.48211,-0.80027 2.81967,-2.2155 2.60437,-3.87561 0.41431,-2.09644 2.12272,-3.84477 2.08912,-6.04389 0.14584,-1.65381 0.0395,-3.32104 -0.24598,-4.95809 2.68336,0.23573 5.75407,-0.78315 6.94353,-3.1182 -2.08971,-1.76923 -5.27004,-2.39848 -7.97287,-1.54619 -1.02473,-1.30279 0.46813,-2.81934 0.84656,-4.12578 0.91633,-1.75813 2.13072,-3.36791 3.33637,-4.96903 z m -4.23542,3.63561 c -0.96958,1.25221 -0.9866,2.91554 -2.01959,4.12969 -1.99164,-0.27206 0.35949,-2.2509 0.85187,-3.05803 0.3584,-0.3836 0.75006,-0.74132 1.16772,-1.07166 z m -10.2251,0.005 c 1.35804,1.07295 2.44056,2.45953 2.98536,4.02539 -2.1235,0.57483 -1.55091,-2.24838 -2.54239,-3.2022 -0.14142,-0.27712 -0.2894,-0.55147 -0.44297,-0.82319 z m 7.47212,5.2867 c 1.78384,-0.43399 0.76968,2.39786 2.33264,1.77605 1.43702,-0.62165 3.16074,-0.85972 4.7153,-0.46689 -0.0893,0.22379 -2.06117,0.35577 -2.82742,0.74232 -1.40552,0.21503 -2.85059,1.44511 -1.81659,2.75031 0.0507,1.53925 -2.52447,1.6417 -2.51325,3.28279 -0.75473,0.85833 -0.31533,3.17945 1.14454,2.03373 0.0798,-0.73015 2.30392,-0.64221 1.4878,0.42087 -0.36189,2.01774 -2.13419,3.6667 -2.00001,5.78783 -0.31426,1.57286 -2.63556,3.43084 -4.23774,2.04668 -1.66643,-0.8304 -1.44565,-2.585 -1.884,-3.99356 -0.54222,-1.41527 -1.60753,-2.691 -1.68838,-4.23185 -0.15132,-1.144 1.81417,-0.14391 2.17244,0.31063 1.88516,-0.0952 0.61247,-2.26558 0.25035,-3.1991 -0.76888,-1.19006 -3.08978,-1.58467 -2.00345,-3.34656 0.23595,-1.80517 -2.62295,-1.88456 -3.9968,-2.34277 -1.7113,-0.4789 1.14108,-0.56295 1.78241,-0.35523 0.93129,0.0332 2.12608,0.75884 2.82336,0.60517 0.162,-0.54336 0.324,-1.08672 0.486,-1.6301 1.34377,-0.56966 2.57706,0.32841 2.85727,1.51853 1.00906,1.7406 1.20773,-1.23902 2.29033,-1.55029 0.18883,-0.1025 0.40658,-0.15566 0.6252,-0.15856 z m 7.6619,2.67555 c -0.4,0.77687 -2.9877,1.01334 -2.96046,0.66857 0.89806,-0.45486 1.96642,-0.51152 2.96046,-0.66857 z m -20.00311,3.1e-4 c 0.95349,0.12487 2.87166,0.40439 3.14705,0.84829 -0.98838,-0.0173 -2.57617,-0.0515 -3.14705,-0.84829 z m 5.53765,4.90309 c 1.38793,0.4695 -0.3547,1.33658 -0.26626,0.0612 z m 9.13856,0.0175 c 0.65249,0.78217 -1.45893,0.89339 -0.429,0.0615 l 0.18714,-0.0454 z m -3.0314,8.30708 c -1.30295,-0.29819 -1.83794,2.64307 -0.31107,1.30643 0.41209,-0.20315 1.37714,-1.31263 0.31107,-1.30643 z m -3.14595,0.0128 c -2.06054,0.40752 2.25846,2.85345 0.89907,0.65083 -0.15921,-0.33003 -0.4893,-0.61776 -0.89907,-0.65083 z"
	SVGIcons["shelf"] = "M 6.5820312 4.5820312 L 6.5839844 26.136719 L 9 26.136719 L 9 22.966797 L 22.998047 22.96875 L 22.998047 26.136719 L 25.416016 26.136719 L 25.416016 4.5820312 L 6.5820312 4.5820312 z M 7.5800781 5.5820312 L 11.029297 5.5820312 L 11.029297 9.0214844 L 7.5800781 9.0214844 L 7.5800781 5.5820312 z M 12.029297 5.5820312 L 15.498047 5.5820312 L 15.498047 9.0214844 L 12.029297 9.0214844 L 12.029297 5.5820312 z M 16.498047 5.5820312 L 19.964844 5.5820312 L 19.964844 9.0214844 L 16.498047 9.0214844 L 16.498047 5.5820312 z M 20.964844 5.5820312 L 24.414062 5.5820312 L 24.414062 9.0214844 L 20.964844 9.0214844 L 20.964844 5.5820312 z M 7.5800781 10.023438 L 11.029297 10.023438 L 11.029297 13.457031 L 7.5800781 13.457031 L 7.5800781 10.023438 z M 12.029297 10.023438 L 15.498047 10.023438 L 15.498047 13.457031 L 12.029297 13.457031 L 12.029297 10.023438 z M 16.498047 10.023438 L 19.964844 10.023438 L 19.966797 13.457031 L 16.498047 13.457031 L 16.498047 10.023438 z M 20.964844 10.023438 L 24.414062 10.023438 L 24.414062 13.457031 L 20.966797 13.457031 L 20.964844 10.023438 z M 15.498047 14.455078 L 15.498047 17.896484 L 12.029297 17.896484 L 12.029297 14.457031 L 15.498047 14.455078 z M 16.498047 14.455078 L 19.966797 14.455078 L 19.966797 17.896484 L 16.498047 17.896484 L 16.498047 14.455078 z M 20.966797 14.455078 L 24.414062 14.455078 L 24.416016 17.896484 L 20.966797 17.896484 L 20.966797 14.455078 z M 7.5800781 14.457031 L 11.029297 14.457031 L 11.029297 17.896484 L 7.5820312 17.896484 L 7.5800781 14.457031 z M 7.6171875 18.792969 L 11.064453 18.792969 L 11.064453 22.234375 L 7.6171875 22.234375 L 7.6171875 18.792969 z M 12.064453 18.792969 L 15.533203 18.792969 L 15.533203 22.234375 L 12.064453 22.234375 L 12.064453 18.792969 z M 16.533203 18.792969 L 20.001953 18.792969 L 20.001953 22.234375 L 16.533203 22.234375 L 16.533203 18.792969 z M 21.001953 18.792969 L 24.451172 18.792969 L 24.451172 22.234375 L 21.001953 22.234375 L 21.001953 18.792969 z"
	SVGIcons["shelf2"] = "M 6.5820312 4.5820312 L 6.5839844 26.136719 L 9 26.136719 L 9 22.966797 L 22.998047 22.96875 L 22.998047 26.136719 L 25.416016 26.136719 L 25.416016 4.5820312 L 6.5820312 4.5820312 z M 8.9492188 5.3925781 L 12.396484 5.3925781 L 12.396484 8.8320312 L 8.9492188 8.8320312 L 8.9492188 5.3925781 z M 14.21875 5.3925781 L 17.6875 5.3925781 L 17.6875 8.8320312 L 14.21875 8.8320312 L 14.21875 5.3925781 z M 19.445312 5.3925781 L 22.912109 5.3925781 L 22.912109 8.8320312 L 19.445312 8.8320312 L 19.445312 5.3925781 z M 8.9492188 9.8339844 L 12.396484 9.8339844 L 12.398438 13.267578 L 8.9492188 13.267578 L 8.9492188 9.8339844 z M 14.21875 9.8339844 L 17.6875 9.8339844 L 17.6875 13.267578 L 14.21875 13.267578 L 14.21875 9.8339844 z M 19.445312 9.8339844 L 22.912109 9.8339844 L 22.912109 13.267578 L 19.445312 13.267578 L 19.445312 9.8339844 z M 19.445312 14.265625 L 22.912109 14.265625 L 22.914062 17.707031 L 19.445312 17.707031 L 19.445312 14.265625 z M 8.9492188 14.267578 L 12.398438 14.267578 L 12.398438 17.707031 L 8.9492188 17.707031 L 8.9492188 14.267578 z M 14.21875 14.267578 L 17.6875 14.267578 L 17.6875 17.707031 L 14.21875 17.707031 L 14.21875 14.267578 z M 8.9863281 18.603516 L 12.433594 18.603516 L 12.433594 22.044922 L 8.9863281 22.044922 L 8.9863281 18.603516 z M 14.253906 18.603516 L 17.722656 18.603516 L 17.722656 22.044922 L 14.253906 22.044922 L 14.253906 18.603516 z M 19.480469 18.603516 L 22.949219 18.603516 L 22.949219 22.044922 L 19.480469 22.044922 L 19.480469 18.603516 z"
	SVGIcons["shelf3"] = "M 6.5820312 4.5820312 L 6.5839844 26.136719 L 9 26.136719 L 9 22.966797 L 22.998047 22.96875 L 22.998047 26.136719 L 25.416016 26.136719 L 25.416016 4.5820312 L 6.5820312 4.5820312 z M 8.9492188 5.3925781 L 15.443359 5.3925781 L 15.443359 8.8320312 L 8.9492188 8.8320312 L 8.9492188 5.3925781 z M 16.230469 5.3925781 L 22.878906 5.3925781 L 22.878906 8.8320312 L 16.230469 8.8320312 L 16.230469 5.3925781 z M 8.9492188 9.8339844 L 15.443359 9.8339844 L 15.447266 13.267578 L 8.9492188 13.267578 L 8.9492188 9.8339844 z M 16.230469 9.8339844 L 22.878906 9.8339844 L 22.878906 13.267578 L 16.230469 13.267578 L 16.230469 9.8339844 z M 16.230469 14.265625 L 22.878906 14.265625 L 22.882812 17.707031 L 16.230469 17.707031 L 16.230469 14.265625 z M 8.9492188 14.267578 L 15.447266 14.267578 L 15.447266 17.707031 L 8.9492188 17.707031 L 8.9492188 14.267578 z M 9.0195312 18.603516 L 15.513672 18.603516 L 15.513672 22.044922 L 9.0195312 22.044922 L 9.0195312 18.603516 z M 16.298828 18.603516 L 22.949219 18.603516 L 22.949219 22.044922 L 16.298828 22.044922 L 16.298828 18.603516 z"
	SVGIcons["shelf4"] = "M 6.5820312 4.5820312 L 6.5839844 26.136719 L 9 26.136719 L 9 22.966797 L 22.998047 22.96875 L 22.998047 26.136719 L 25.416016 26.136719 L 25.416016 4.5820312 L 22.878906 4.5820312 L 22.878906 8.0292969 L 16.230469 8.0292969 L 16.230469 4.5820312 L 15.443359 4.5820312 L 15.443359 8.0292969 L 8.9492188 8.0292969 L 8.9492188 4.5820312 L 6.5820312 4.5820312 z M 8.9492188 9.03125 L 15.443359 9.03125 L 15.447266 12.464844 L 8.9492188 12.464844 L 8.9492188 9.03125 z M 16.230469 9.03125 L 22.878906 9.03125 L 22.878906 12.464844 L 16.230469 12.464844 L 16.230469 9.03125 z M 16.230469 13.462891 L 22.878906 13.462891 L 22.882812 16.904297 L 16.230469 16.904297 L 16.230469 13.462891 z M 8.9492188 13.464844 L 15.447266 13.464844 L 15.447266 16.904297 L 8.9492188 16.904297 L 8.9492188 13.464844 z M 9.0195312 17.800781 L 15.513672 17.800781 L 15.513672 21.242188 L 9.0195312 21.242188 L 9.0195312 17.800781 z M 16.298828 17.800781 L 22.949219 17.800781 L 22.949219 21.242188 L 16.298828 21.242188 L 16.298828 17.800781 z"
	SVGIcons["sorting1"] = "m 22.519115,30.708898 c -0.123849,-0.12385 -0.16975,-1.22841 -0.16975,-4.08498 0,-4.1475 0.02879,-4.32962 0.665133,-4.20723 0.235786,0.0454 0.257104,0.31462 0.292101,3.68961 l 0.03775,3.64068 h 3.460961 3.460961 l 0.03775,-3.64068 c 0.035,-3.37499 0.05631,-3.64426 0.292101,-3.68961 0.636342,-0.12239 0.665133,0.0597 0.665133,4.20723 0,2.85657 -0.0459,3.96113 -0.16975,4.08498 -0.242378,0.24238 -8.330015,0.24238 -8.572394,0 z M 1.3003209,30.567438 c -0.1238506,-0.12385 -0.1697504,-1.22841 -0.1697504,-4.08498 0,-4.1475 0.028791,-4.32962 0.665133,-4.20723 0.2357855,0.0454 0.2571033,0.31462 0.2921009,3.68961 l 0.037753,3.64068 h 3.4609601 3.4609604 l 0.03775,-3.64068 c 0.035,-3.37499 0.05632,-3.64426 0.2921,-3.68961 0.6363421,-0.12239 0.6651331,0.0597 0.6651331,4.20723 0,2.85657 -0.0459,3.96113 -0.1697501,4.08498 -0.242379,0.24238 -8.3300143,0.24238 -8.5723929,0 z m 10.6093971,0 c -0.12385,-0.12385 -0.16975,-1.22841 -0.16975,-4.08498 0,-4.1475 0.02879,-4.32962 0.665133,-4.20723 0.235786,0.0454 0.257104,0.31462 0.292101,3.68961 l 0.03775,3.64068 h 3.460961 3.46096 l 0.03775,-3.64068 c 0.035,-3.37499 0.05631,-3.64426 0.292101,-3.68961 0.636341,-0.12239 0.665133,0.0597 0.665133,4.20723 0,2.85657 -0.0459,3.96113 -0.169751,4.08498 -0.242378,0.24238 -8.330015,0.24238 -8.572393,0 z m 1.386295,-3.86182 v -2.47552 h 2.899902 2.899902 v 2.47552 2.47553 h -2.899902 -2.899902 z m 10.467938,1.34386 v -1.13167 h 2.970632 2.970631 v 1.13167 1.13167 H 26.734583 23.763951 Z M 2.5451568,27.200728 v -1.69751 h 2.9706313 2.9706318 v 1.69751 1.6975 H 5.5157881 2.5451568 Z m 12.2682172,-5.90903 -1.092985,-1.09943 v -0.81026 c 0,-0.44564 0.04533,-0.81026 0.100739,-0.81026 0.05541,0 0.373689,0.26788 0.707293,0.59529 l 0.606554,0.59529 v -3.42447 -3.42445 h 0.848752 0.848751 v 3.42445 3.42447 l 0.606554,-0.59529 c 0.333605,-0.32741 0.651886,-0.59529 0.707293,-0.59529 0.05541,0 0.10074,0.40161 0.10074,0.89247 v 0.89248 l -1.121088,1.01722 c -0.6166,0.55947 -1.143258,1.01721 -1.170353,1.01721 -0.0271,0 -0.541107,-0.49474 -1.14225,-1.09943 z m -4.487992,-6.18569 v -0.77802 H 9.5473589 8.7693369 v -0.63656 -0.63657 h 0.778022 0.7780231 v -0.77802 -0.77802 h 0.636564 0.636563 v 0.77802 0.77802 h 0.778023 0.778022 v 0.63657 0.63656 h -0.778022 -0.778023 v 0.77802 0.77803 h -0.636563 -0.636564 z m 4.243758,-4.73886 V 9.0940178 H 7.8357099 1.1022788 l -0.48095926,-0.48096 c -0.3433708,-0.34337 -0.4809594,-0.60841 -0.4809594,-0.9265 0,-0.60389 0.3252339,-1.18967 0.7740698,-1.39417 0.53542076,-0.24395 29.74463106,-0.24395 30.28005206,0 0.448837,0.2045 0.77407,0.79028 0.77407,1.39417 0,0.31809 -0.137588,0.58313 -0.480959,0.9265 l -0.48096,0.48096 h -6.73343 -6.733431 v 1.2731302 c 0,1.03736 -0.0393,1.27313 -0.212188,1.27313 -0.172894,0 -0.212188,-0.23577 -0.212188,-1.27313 V 9.0940178 h -1.06094 -1.06094 v 1.2731302 c 0,1.03736 -0.03929,1.27313 -0.212188,1.27313 -0.172893,0 -0.212188,-0.23577 -0.212188,-1.27313 z M 2.2177578,8.3620678 c 0.580767,-0.5456 0.3915949,-1.49474 -0.3599613,-1.80605 -0.7602084,-0.31489 -1.61520846,0.43013 -1.45159426,1.26487 0.1525661,0.77838 1.21972426,1.09718 1.81155556,0.54118 z m -1.1343401,-0.35256 c -0.19513336,-0.19514 -0.083982,-0.75002 0.1783639,-0.89042 0.5227947,-0.2798 0.9871336,0.22509 0.6920142,0.75244 -0.1289125,0.23036 -0.6891832,0.31917 -0.8703781,0.13798 z m 10.4485233,0.56824 c 0.131718,-0.0506 0.320592,-0.28479 0.419718,-0.52055 0.241562,-0.5745 0.02719,-1.16887 -0.524471,-1.45415 -1.140962,-0.59001 -2.1657981,1.11015 -1.139129,1.88977 0.251692,0.19113 0.859058,0.2326 1.243882,0.0849 z m -1.039574,-0.70622 c -0.280438,-0.50111 0.180611,-1.02854 0.718628,-0.82208 0.531051,0.20378 0.330137,1.05436 -0.249049,1.05436 -0.186774,0 -0.398084,-0.10452 -0.469579,-0.23228 z m 11.64897,0.70622 c 0.13172,-0.0506 0.320592,-0.28479 0.41972,-0.52055 0.241561,-0.5745 0.02719,-1.16887 -0.524472,-1.45415 -1.140962,-0.59001 -2.165799,1.11015 -1.139129,1.88977 0.251692,0.19113 0.859059,0.2326 1.243881,0.0849 z m -1.039573,-0.70622 c -0.280438,-0.50111 0.180611,-1.02854 0.718628,-0.82208 0.531051,0.20378 0.330136,1.05436 -0.249049,1.05436 -0.186774,0 -0.398085,-0.10452 -0.469579,-0.23228 z m 10.264039,0.47983 c 0.547351,-0.58991 0.427531,-1.38457 -0.262579,-1.74144 -1.086756,-0.56198 -2.0779,0.83179 -1.231894,1.73233 0.405359,0.43148 1.098656,0.43571 1.494473,0.009 z m -1.206058,-0.47156 c -0.301904,-0.53947 0.160956,-1.04245 0.695863,-0.75618 0.434876,0.23274 0.306839,0.89635 -0.183891,0.9531 -0.242116,0.028 -0.425598,-0.0426 -0.511972,-0.19692 z"
	SVGIcons["store1"] = "M 10.85826,18.866847 10.337739,18.533198 V 16.328707 14.124215 H 8.5243116 6.7108847 v 1.665123 1.665125 L 6.2575283,17.416726 C 6.0081819,17.396016 5.7361678,17.378857 5.6530525,17.378593 L 5.5019336,17.37817 v -1.956272 c 0,-1.235371 0.025621,-1.998442 0.069527,-2.07071 0.067062,-0.110374 0.170552,-0.114431 2.9192694,-0.114431 2.748717,0 2.852206,0.004 2.919268,0.114431 0.08096,0.133247 0.0997,5.86456 0.01916,5.85516 -0.0277,-0.0033 -0.284606,-0.156016 -0.570892,-0.339515 z m 11.061062,4.538636 v -2.216408 h 3.055959 3.055961 v 2.216408 2.216415 h -3.055961 -3.055959 z m -10.960341,3.610359 c -0.04619,-0.02936 -0.113281,-0.28443 -0.149093,-0.566891 -0.08955,-0.706266 -0.121788,-0.8212 -0.381868,-1.361288 -0.5343282,-1.109592 -1.5096828,-1.925602 -2.6610036,-2.226278 -0.54296,-0.141799 -1.5053787,-0.140938 -2.0522246,0.0015 -0.2338597,0.06104 -0.4529821,0.111011 -0.4869387,0.111011 -0.036488,0 -0.061739,-1.071343 -0.061739,-2.619397 v -2.619398 h 0.7028401 c 2.6396471,0 5.2916518,1.446293 6.7929628,3.704582 l 0.305531,0.459589 h 2.412745 2.412744 v 1.544766 1.544774 h -1.028851 c -1.117475,0 -1.459908,0.06434 -2.035784,0.382723 -0.473997,0.262031 -1.042034,0.86661 -1.236907,1.316478 l -0.164861,0.380589 -1.141787,1.34e-4 c -0.627983,6.8e-5 -1.179577,-0.02389 -1.225766,-0.05318 z m 9.309964,-3.860911 v -3.941889 h 0.305826 c 0.232735,0 0.361149,0.04351 0.53731,0.182187 l 0.231483,0.182188 2e-6,3.329075 2e-6,3.329074 h 2.88805 2.888051 v 0.430627 0.43062 h -3.425363 -3.425361 z m -1.239407,3.758911 C 18.963098,26.74985 18.778637,26.449045 18.61961,26.245394 l -0.289141,-0.370277 v -6.29995 -6.299953 h 0.671936 0.671936 l -0.01708,6.949139 -0.01708,6.949138 -0.244921,0.01928 c -0.235793,0.01857 -0.249419,0.0081 -0.365708,-0.278911 z M 7.3813093,28.0793 c 0.4523978,-0.230799 0.7592082,-0.732119 0.7592082,-1.240533 0,-0.849915 -0.7790525,-1.539313 -1.6003259,-1.416155 -0.8011376,0.120139 -1.3832519,1.000581 -1.1743854,1.776252 0.089026,0.33062 0.4003947,0.704147 0.733909,0.880421 0.3321564,0.175562 0.9374849,0.175569 1.2815941,1.5e-5 z M 5.821765,29.631084 C 4.0436147,29.04546 3.231809,27.113506 4.0546713,25.425704 5.0735329,23.335886 8.0087036,23.168636 9.2664945,25.128737 10.278928,26.706472 9.586047,28.896178 7.862917,29.564441 7.3251451,29.773003 6.3494087,29.804856 5.821765,29.631091 Z m 10.831112,-1.073651 c 0.528734,-0.528733 0.213739,-1.351888 -0.517326,-1.351888 -0.255949,0 -0.339772,0.03595 -0.532159,0.228365 -0.19151,0.191507 -0.228358,0.276745 -0.228358,0.528253 0,0.494333 0.296532,0.775101 0.827925,0.783902 0.196894,0.003 0.302162,-0.04089 0.449918,-0.188632 z m -1.163799,1.174565 c -0.961796,-0.358248 -1.473121,-1.317376 -1.220131,-2.288674 0.274279,-1.053024 1.227832,-1.642251 2.300092,-1.421286 0.497897,0.102607 0.887303,0.361252 1.157081,0.76854 0.269338,0.406624 0.356938,0.759869 0.319432,1.2881 -0.04891,0.688784 -0.415864,1.249684 -1.021447,1.561288 -0.348717,0.179435 -1.168427,0.228574 -1.535027,0.09203 z"
	SVGIcons["stock1"] = "M 3.0770447,24.788513 V 18.619492 L 9.479796,12.206494 c 5.129599,-5.1378039 6.448961,-6.4129887 6.635139,-6.4129887 0.186177,0 1.50554,1.2751848 6.635138,6.4129887 l 6.402751,6.412998 v 6.169021 6.16903 H 28.104323 27.055821 V 25.236085 19.514627 L 21.585378,14.044756 16.114935,8.5748846 10.644491,14.044756 5.1740478,19.514627 v 5.721458 5.721458 H 4.1255467 3.0770447 Z m 6.491593,6.059618 c -0.1521919,-0.152192 -0.1521919,-2.546555 0,-2.698748 0.152191,-0.152193 2.5465603,-0.152193 2.6987523,0 0.152191,0.152193 0.152191,2.546556 0,2.698748 -0.152192,0.152193 -2.5465613,0.152193 -2.6987523,0 z m 3.4646133,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349378 0,-0.842851 0.03504,-1.275001 0.109409,-1.34937 0.152191,-0.152193 2.546561,-0.152193 2.698752,0 0.152191,0.152193 0.152191,2.546556 0,2.698748 -0.152191,0.152193 -2.546561,0.152193 -2.698752,0 z m 6.929229,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349378 0,-0.842851 0.03504,-1.275001 0.109409,-1.34937 0.152191,-0.152193 2.54656,-0.152193 2.698752,0 0.152191,0.152193 0.152191,2.546556 0,2.698748 -0.152192,0.152193 -2.546561,0.152193 -2.698752,0 z m 3.464614,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349378 0,-0.842851 0.03504,-1.275001 0.109409,-1.34937 0.152192,-0.152193 2.546561,-0.152193 2.698752,0 0.152192,0.152193 0.152192,2.546556 0,2.698748 -0.152191,0.152193 -2.54656,0.152193 -2.698752,0 z M 6.1040233,27.38352 c -0.152191,-0.152193 -0.152191,-2.546565 0,-2.698757 0.1521919,-0.152184 2.5465607,-0.152184 2.6987517,0 0.1521919,0.152192 0.1521919,2.546564 0,2.698757 -0.152191,0.152193 -2.5465598,0.152193 -2.6987517,0 z m 3.4646144,0 c -0.1521919,-0.152193 -0.1521919,-2.546565 0,-2.698757 0.152191,-0.152184 2.5465603,-0.152184 2.6987523,0 0.152191,0.152192 0.152191,2.546564 0,2.698757 -0.152192,0.152193 -2.5465613,0.152193 -2.6987523,0 z m 3.4646133,0 c -0.07437,-0.07435 -0.109409,-0.506528 -0.109409,-1.349379 0,-0.84285 0.03504,-1.275001 0.109409,-1.349378 0.152191,-0.152184 2.546561,-0.152184 2.698752,0 0.152191,0.152192 0.152191,2.546564 0,2.698757 -0.152191,0.152193 -2.546561,0.152193 -2.698752,0 z m 3.464614,0 c -0.07437,-0.07435 -0.109409,-0.506528 -0.109409,-1.349379 0,-0.84285 0.03504,-1.275001 0.109409,-1.349378 0.152192,-0.152184 2.546561,-0.152184 2.698752,0 0.152192,0.152192 0.152192,2.546564 0,2.698757 -0.152191,0.152193 -2.54656,0.152193 -2.698752,0 z m 3.464615,0 c -0.07437,-0.07435 -0.109409,-0.506528 -0.109409,-1.349379 0,-0.84285 0.03504,-1.275001 0.109409,-1.349378 0.152191,-0.152184 2.54656,-0.152184 2.698752,0 0.152191,0.152192 0.152191,2.546564 0,2.698757 -0.152192,0.152193 -2.546561,0.152193 -2.698752,0 z m 3.464614,0 c -0.07437,-0.07435 -0.109409,-0.506528 -0.109409,-1.349379 0,-0.84285 0.03504,-1.275001 0.109409,-1.349378 0.152192,-0.152184 2.546561,-0.152184 2.698752,0 0.152192,0.152192 0.152192,2.546564 0,2.698757 -0.152191,0.152193 -2.54656,0.152193 -2.698752,0 z M 6.1040233,24.010078 c -0.152191,-0.152193 -0.152191,-2.546565 0,-2.698749 0.1521919,-0.152192 2.5465607,-0.152192 2.6987517,0 0.1521919,0.152184 0.1521919,2.546556 0,2.698749 -0.074369,0.07435 -0.5065227,0.109411 -1.349376,0.109411 -0.8428533,0 -1.2750062,-0.03501 -1.3493757,-0.109411 z m 3.4646144,0 c -0.1521919,-0.152193 -0.1521919,-2.546565 0,-2.698749 0.152191,-0.152192 2.5465603,-0.152192 2.6987523,0 0.152191,0.152184 0.152191,2.546556 0,2.698749 -0.07437,0.07435 -0.506524,0.109411 -1.349377,0.109411 -0.842853,0 -1.2750067,-0.03501 -1.3493753,-0.109411 z m 6.9292273,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349379 0,-0.842851 0.03504,-1.275001 0.109409,-1.34937 0.152192,-0.152192 2.546561,-0.152192 2.698752,0 0.152192,0.152184 0.152192,2.546556 0,2.698749 -0.07437,0.07435 -0.506523,0.109411 -1.349376,0.109411 -0.842852,0 -1.275005,-0.03501 -1.349376,-0.109411 z m 3.464615,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349379 0,-0.842851 0.03504,-1.275001 0.109409,-1.34937 0.152191,-0.152192 2.54656,-0.152192 2.698752,0 0.152191,0.152184 0.152191,2.546556 0,2.698749 -0.07437,0.07435 -0.506523,0.109411 -1.349376,0.109411 -0.842853,0 -1.275007,-0.03501 -1.349376,-0.109411 z m 3.464614,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349379 0,-0.842851 0.03504,-1.275001 0.109409,-1.34937 0.152192,-0.152192 2.546561,-0.152192 2.698752,0 0.152192,0.152184 0.152192,2.546556 0,2.698749 -0.07437,0.07435 -0.506523,0.109411 -1.349376,0.109411 -0.842854,0 -1.275007,-0.03501 -1.349376,-0.109411 z M 9.5686377,20.636635 c -0.1521919,-0.152192 -0.1521919,-2.546555 0,-2.698748 0.152191,-0.152193 2.5465603,-0.152193 2.6987523,0 0.152191,0.152193 0.152191,2.546556 0,2.698748 -0.07437,0.07435 -0.506524,0.109412 -1.349377,0.109412 -0.842853,0 -1.2750067,-0.03501 -1.3493753,-0.109412 z m 3.4646133,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349369 0,-0.84286 0.03504,-1.27501 0.109409,-1.349379 0.152191,-0.152193 2.546561,-0.152193 2.698752,0 0.152191,0.152193 0.152191,2.546556 0,2.698748 -0.07437,0.07435 -0.506522,0.109412 -1.349376,0.109412 -0.842853,0 -1.275006,-0.03501 -1.349376,-0.109412 z m 3.464614,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349369 0,-0.84286 0.03504,-1.27501 0.109409,-1.349379 0.152192,-0.152193 2.546561,-0.152193 2.698752,0 0.152192,0.152193 0.152192,2.546556 0,2.698748 -0.07437,0.07435 -0.506523,0.109412 -1.349376,0.109412 -0.842852,0 -1.275005,-0.03501 -1.349376,-0.109412 z m 3.464615,0 c -0.07437,-0.07435 -0.109409,-0.506519 -0.109409,-1.349369 0,-0.84286 0.03504,-1.27501 0.109409,-1.349379 0.152191,-0.152193 2.54656,-0.152193 2.698752,0 0.152191,0.152193 0.152191,2.546556 0,2.698748 -0.07437,0.07435 -0.506523,0.109412 -1.349376,0.109412 -0.842853,0 -1.275007,-0.03501 -1.349376,-0.109412 z"

	SVGIcons["trend"] = "m 27.81612,7.6892051 -6,1.221 1.716,1.7079999 -5.351,5.358 -3.001,-3.002 -7.336,7.242 1.41,1.418 5.922,-5.834 2.991,2.993 6.781,-6.762 1.667,1.66 1.201,-6.0019999 z m -16.69,6.4769999 -3.282,-3.239 1.41,-1.4179999 3.298,3.2489999 -1.426,1.408 z m 15.49,3.287 1.2,6.001 -6,-1.221 1.716,-1.708 -2.13,-2.133 1.411,-1.408 2.136,2.129 1.667,-1.66 z m 1.2,8.181 0,2 -24,0 0,-21.9999999 2,0 0,19.9999999 22,0 z"
	SVGIcons["trend_up"] = "m 27.81612,8.5092052 -6,1.221 1.716,1.7079998 -5.351,5.358 -3.001,-3.002 -7.336,7.242 1.41,1.418 5.922,-5.834 2.991,2.993 6.781,-6.762 1.667,1.66 1.201,-6.0019998 z m 0,17.1249998 0,2 -24,0 0,-21.9999998 2,0 0,19.9999998 22,0 z"
	SVGIcons["trend_down"] = "m 26.61612,16.435105 -1.667,1.826 -6.781,-7.4382 -2.992,3.2923 -5.922,-6.4184999 -1.41,1.5598 7.336,7.9661999 3.001,-3.3022 5.351,5.8938 -1.716,1.8788 6,1.3431 -1.2,-6.6011 z m 1.2,8.9991 0,2.2 -24,0 0,-21.9999999 2,0 0,19.7999999 22,0 z"
	SVGIcons["truck"] = "m 6.8161139,22.634205 -2,0 c -0.552,0 -1,-0.448 -1,-1 l 0,-2 15.0000061,0 0,-9 4.667,0 c 1.117,0 1.6,0.576 1.936,1.107 0.594,0.94 1.536,2.432 2.109,3.378 0.188,0.312 0.288,0.67 0.288,1.035 l 0,4.48 c 0,1.121 -0.728,2 -2,2 l -1,0 c 0,1.656 -1.344,3 -3,3 -1.656,0 -3,-1.344 -3,-3 l -6,0 c 0,1.656 -1.344,3 -3.000006,3 -1.6560001,0 -3.0000001,-1.344 -3.0000001,-3 z m 3.0000001,-1.2 c 0.662004,0 1.200006,0.538 1.200006,1.2 0,0.662 -0.538002,1.2 -1.200006,1.2 -0.662,0 -1.2000001,-0.538 -1.2000001,-1.2 0,-0.662 0.5380001,-1.2 1.2000001,-1.2 z m 12.000006,0 c 0.662,0 1.2,0.538 1.2,1.2 0,0.662 -0.538,1.2 -1.2,1.2 -0.662,0 -1.2,-0.538 -1.2,-1.2 0,-0.662 0.538,-1.2 1.2,-1.2 z m -4,-2.8 -14.0000061,0 0,-10.0000003 c 0,-0.552 0.448,-1 1,-1 l 12.0000061,0 c 0.552,0 1,0.448 1,1 l 0,10.0000003 z m 3,-6 0,3 4.715,0 -1.427,-2.496 c -0.178,-0.312 -0.509,-0.504 -0.868,-0.504 l -2.42,0 z"
	SVGIcons["truck2"] = "m 6.81612,22.634205 -2,0 c -0.552,0 -1,-0.448 -1,-1 l 0,-13 c 0,-0.552 0.448,-1 1,-1 l 13,0 c 0.552,0 1,0.448 1,1 l 0,2 4.667,0 c 1.117,0 1.6,0.576 1.936,1.107 0.594,0.94 1.536,2.432 2.109,3.378 0.188,0.312 0.288,0.67 0.288,1.035 l 0,4.48 c 0,1.089 -0.743,2 -2,2 l -1,0 c 0,1.656 -1.344,3 -3,3 -1.656,0 -3,-1.344 -3,-3 l -6,0 c 0,1.656 -1.344,3 -3,3 -1.656,0 -3,-1.344 -3,-3 z m 3,-1.2 c 0.662,0 1.2,0.538 1.2,1.2 0,0.662 -0.538,1.2 -1.2,1.2 -0.662,0 -1.2,-0.538 -1.2,-1.2 0,-0.662 0.538,-1.2 1.2,-1.2 z m 12,0 c 0.662,0 1.2,0.538 1.2,1.2 0,0.662 -0.538,1.2 -1.2,1.2 -0.662,0 -1.2,-0.538 -1.2,-1.2 0,-0.662 0.538,-1.2 1.2,-1.2 z m -3,-2.8 -13,0 0,2 1.765,0 c 0.549,-0.614 1.347,-1 2.235,-1 0.888,0 1.686,0.386 2.235,1 l 7.53,0 c 0.549,-0.614 1.347,-1 2.235,-1 0.888,0 1.686,0.386 2.235,1 l 1.765,0 0,-4.575 -1.711,-2.929 c -0.179,-0.307 -0.508,-0.496 -0.863,-0.496 l -4.426,0 0,6 z m -2,-9 -11,0 0,7 11,0 0,-7 z m 3,4 0,3 5,0 -1.427,-2.496 c -0.178,-0.312 -0.509,-0.504 -0.868,-0.504 l -2.705,0 z"
	SVGIcons["tube"] = "m 17.033668,8.6589983 2.010522,0 0,16.0841837 c 0,2.220622 -1.800422,4.021045 -4.021044,4.021045 -2.220622,0 -4.021045,-1.800423 -4.021045,-4.021045 l 0,-16.0841837 2.010523,0 0,3.0157877 3.015783,0 0,12.890465 c 0,0.978119 -0.329725,1.700901 -1.019335,2.186443 l 0.01407,0.0021 c 1.108804,0 2.010523,-0.901719 2.010523,-2.010522 l 0,-16.084184 z m 2.010522,-4.0210449 -8.042089,0 c -0.554904,0 -1.0052645,0.4503571 -1.0052645,1.0052612 0,0.5549043 0.4503605,1.0052613 1.0052645,1.0052613 l 8.042089,0 c 0.55591,0 1.005262,-0.450357 1.005262,-1.0052613 0,-0.5549041 -0.449352,-1.0052612 -1.005262,-1.0052612 z"
	SVGIcons["tube2"] = "m 17.07461,24.791346 c 0,1.111704 -0.904079,2.015783 -2.015784,2.015783 l -0.01411,-0.001 c 0.691413,-0.486811 1.022002,-1.212493 1.022002,-2.193172 l 0,-12.924197 -3.023676,0 0,-3.023669 -2.015783,0 0,16.126263 c 0,2.226433 1.805134,4.031567 4.031567,4.031567 2.226433,0 4.031567,-1.805134 4.031567,-4.031567 l 0,-16.126263 -2.015783,0 0,16.126263 z M 13.043042,14.712428 c 0.557365,0 1.007892,0.451535 1.007892,1.007891 0,0.556357 -0.450527,1.007892 -1.007892,1.007892 -0.557364,0 -1.007891,-0.451535 -1.007891,-1.007892 0,-0.556356 0.450527,-1.007891 1.007891,-1.007891 z m 0,10.078918 c -0.557364,0 -1.007891,-0.451536 -1.007891,-1.007892 0,-0.556356 0.450527,-1.007892 1.007891,-1.007892 0.557365,0 1.007892,0.451536 1.007892,1.007892 0,0.556356 -0.450527,1.007892 -1.007892,1.007892 z m 1.007892,-4.031568 c -0.557364,0 -1.007892,-0.451535 -1.007892,-1.007891 0,-0.556357 0.450528,-1.007892 1.007892,-1.007892 0.557364,0 1.007892,0.451535 1.007892,1.007892 0,0.556356 -0.450528,1.007891 -1.007892,1.007891 z M 20.098285,5.6414077 c 0,0.5563563 -0.450527,1.0078918 -1.007892,1.0078918 l -8.063134,0 c -0.556358,0 -1.007898,-0.4515355 -1.007898,-1.0078918 0,-0.5563563 0.45154,-1.0078918 1.007898,-1.0078918 l 8.063134,0 c 0.557365,0 1.007892,0.4515355 1.007892,1.0078918 z"
	SVGIcons["tube3"] = "m 20.098,5.6420797 c 0,0.556347 -0.450531,1.007875 -1.0079,1.007875 l -8.0632,0 c -0.556361,0 -1.0079,-0.451528 -1.0079,-1.007875 0,-0.556347 0.451539,-1.007875 1.0079,-1.007875 l 8.0632,0 c 0.557369,0 1.0079,0.451528 1.0079,1.007875 z M 17.0743,24.791705 c 0,1.111686 -0.904086,2.01575 -2.0158,2.01575 -1.111714,0 -2.0158,-0.904064 -2.0158,-2.01575 l 0,-3.023625 2.0158,0 0,-1.007875 -2.0158,0 0,-2.01575 2.0158,0 0,-1.007875 -2.0158,0 0,-2.01575 2.0158,0 0,-1.007875 -2.0158,0 0,-2.01575 2.0158,0 0,-1.007875 -2.0158,0 0,-2.0157503 2.0158,0 0,-1.007875 -4.0316,0 0,16.1260003 c 0,2.226396 1.805149,4.0315 4.0316,4.0315 2.226451,0 4.0316,-1.805104 4.0316,-4.0315 l 0,-16.1260003 -2.0158,0 0,16.1260003 z"
	SVGIcons["tube4"] = "m 18.953646,12.634205 1.558058,0 0,12 c 0,2.209 -1.395241,4 -3.116117,4 -1.720875,0 -3.116116,-1.791 -3.116116,-4 l 0,-12 1.558058,0 0,3 2.337088,0 0,8.823 c 0,0.973 -0.255522,1.692 -0.789936,2.175 l 0.01091,0.002 c 0.85927,0 1.558059,-0.897 1.558059,-2 l 0,-12 z m 1.558058,-4.0000003 -6.232233,0 c -0.430024,0 -0.779029,0.448 -0.779029,1 0,0.5520003 0.349005,1.0000003 0.779029,1.0000003 l 6.232233,0 c 0.430803,0 0.779029,-0.448 0.779029,-1.0000003 0,-0.552 -0.348226,-1 -0.779029,-1 z m -9.34835,12.0000003 0,-3 1.558059,0 0,-1 -1.558059,0 0,-1 1.558059,0 0,-1 -1.558059,0 0,-1 1.558059,0 0,-1 -1.558059,0 0,-1 1.558059,0 0,-1 -1.558059,0 0,-1.0000003 1.558059,0 0,-1 -3.1161169,0 0,12.0000003 c 0,2.209 1.3952409,4 3.1161169,4 l 0,-2 c -0.85927,0 -1.558059,-0.897 -1.558059,-2 z m 3.116117,-14.0000003 1.558058,0 c 0.430803,0 0.779029,-0.448 0.779029,-1 0,-0.552 -0.348226,-1 -0.779029,-1 l -6.2322329,0 c -0.4300241,0 -0.7790291,0.448 -0.7790291,1 0,0.552 0.349005,1 0.7790291,1 l 4.6741749,0 z"

	SVGIcons["hand_w"] = "m 22.35212,12.189205 c -1.188,-0.252 -4.606,-0.904 -5.536,-1.088 l 0,-3.5119999 c 0,-1.629 -1.346,-2.955 -3,-2.955 -1.654,0 -3,1.326 -3,2.955 l 0,7.4569999 c -0.554,-0.336 -1.188,-0.621 -1.838,-0.715 -1.822,-0.262 -3.162,0.94 -3.162,2.498 0,0.805 0.363,1.613 1.022,2.271 3.972,3.972 5.688,5.125 6.059,9.534 l 9.919,0 0,-1.748 c 0,-5.154 3,-6.031 3,-10.029 0,-2.448 -1.061,-4.157 -3.464,-4.668 z m 0.357,8.022 c -0.821,1.483 -1.838,3.319 -1.891,6.423 l -6.13,0 c -0.726,-3.82 -3.81,-6.318 -6.436,-8.949 -0.688,-0.686 -0.393,-1.37 0.442,-1.373 1.263,-0.006 3.06,1.884 4.122,3.205 l 0,-11.9279999 c 0,-0.517 0.458,-0.955 1,-0.955 0.542,0 1,0.438 1,0.955 l 0,6.9479999 c 0,0.315 0.256,0.571 0.572,0.571 0.314,0 0.57,-0.256 0.57,-0.571 l 0,-0.575 c 0,-0.534 0.49,-0.938 1.014,-0.833 0.398,0.079 0.686,0.428 0.686,0.833 l 0,1.273 c 0,0.315 0.256,0.571 0.571,0.571 0.315,0 0.571,-0.256 0.571,-0.571 l 0,-0.83 c 0,-0.531 0.487,-0.932 1.008,-0.828 0.396,0.078 0.682,0.424 0.682,0.828 l 0,1.533 c 0,0.315 0.256,0.571 0.571,0.571 0.315,0 0.571,-0.256 0.571,-0.571 l 0,-0.912 c 0,-0.523 0.545,-0.867 1.018,-0.646 0.645,0.305 1.166,0.932 1.166,2.477 0,1.355 -0.465,2.193 -1.107,3.354 z"
	SVGIcons["hand_b"] = "m 7.83712,15.322205 c 1.208,0.172 2.51,1.312 2.979,1.781 l 0,-10.5140002 c 0,-1.08 0.92,-1.955 2,-1.955 1.08,0 2,0.875 2,1.955 l 0,6.0580002 c 0,0.784 0.814,0.885 0.919,0.103 0.216,-1.604 2.519,-1.817 2.693,0.399 0.043,0.546 0.726,0.655 0.866,0.027 0.326,-1.444 2.501,-1.458 2.758,0.758 0.066,0.579 0.796,0.696 0.848,0.034 0.051,-0.67 0.281,-0.934 0.607,-0.934 1.098,0 2.309,2.019 2.309,4.41 0,4.295 -3,4.306 -3,11.19 l -10,0 c -0.332,-3.942 -3.462,-7.431 -6.271,-10.241 -0.488,-0.488 -0.729,-1.052 -0.729,-1.564 0,-0.93 0.759,-1.688 2.021,-1.507 z"

	SVGIcons["view2"] = "m 16.81612,16.634205 c 0,1.657 -1.343,3 -3,3 -1.657,0 -3,-1.343 -3,-3 0,-0.199 0.02,-0.393 0.057,-0.581 1.474,0.541 2.927,-0.882 2.405,-2.371 0.174,-0.03 0.354,-0.048 0.538,-0.048 1.657,0 3,1.344 3,3 z m -2.985,-6.9999998 c -7.569,0 -12.015,6.5509998 -12.015,6.5509998 0,0 4.835,7.449 12.015,7.449 7.733,0 11.985,-7.449 11.985,-7.449 0,0 -4.291,-6.5509998 -11.985,-6.5509998 z m -0.015,11.9999998 c -2.761,0 -5,-2.238 -5,-5 0,-2.761 2.239,-5 5,-5 2.762,0 5,2.239 5,5 0,2.762 -2.238,5 -5,5 z"
	SVGIcons["view3"] = "m 15.83112,11.634205 c 4.751,0 8.063,3.012 9.504,4.636 -1.401,1.837 -4.713,5.364 -9.504,5.364 -4.42,0 -7.93,-3.536 -9.478,-5.407 1.493,-1.647 4.817,-4.593 9.478,-4.593 z m 0,-1.9999999 c -7.569,0 -12.015,6.5509999 -12.015,6.5509999 0,0 4.835,7.449 12.015,7.449 7.733,0 11.985,-7.449 11.985,-7.449 0,0 -4.291,-6.5509999 -11.985,-6.5509999 z m -0.015,2.9999999 c -2.21,0 -4,1.791 -4,4 0,2.209 1.79,4 4,4 2.209,0 4,-1.791 4,-4 0,-2.209 -1.791,-4 -4,-4 z m -0.004,3.999 c -0.564,0.564 -1.479,0.564 -2.044,0 -0.565,-0.564 -0.565,-1.48 0,-2.044 0.564,-0.564 1.479,-0.564 2.044,0 0.565,0.564 0.565,1.479 0,2.044 z"
	SVGIcons["view4"] = "m 14.23112,11.634205 c 4.751,0 8.063,3.012 9.504,4.636 -1.401,1.837 -4.713,5.364 -9.504,5.364 -4.42,0 -7.93,-3.536 -9.478,-5.407 1.493,-1.647 4.817,-4.593 9.478,-4.593 z m 0,-2 c -7.569,0 -12.015,6.551 -12.015,6.551 0,0 4.835,7.449 12.015,7.449 7.733,0 11.985,-7.449 11.985,-7.449 0,0 -4.291,-6.551 -11.985,-6.551 z m -0.015,7 -3.36,-2.171 c -0.405,0.625 -0.64,1.371 -0.64,2.171 0,2.209 1.791,4 4,4 2.209,0 4,-1.791 4,-4 0,-2.209 -1.791,-4 -4,-4 -0.742,0 -1.438,0.202 -2.033,0.554 l 2.033,3.446 z"

	SVGIcons["stock"] = "M 3.8029972,21.033894 3.9177538,12.694907 9.7556105,9.1284873 15.593467,5.562068 l 5.837857,3.5664193 5.837857,3.5664197 0.114757,8.338987 0.114757,8.338988 h -1.282329 -1.282328 l -0.0045,-7.556436 -0.0045,-7.556422 -4.665763,-2.998154 -4.665763,-2.9981425 -4.665764,2.9981425 -4.6657626,2.998154 -0.00452,7.556422 -0.00452,7.556436 H 4.9705711 3.6882432 Z m 3.299042,5.920926 v -2.41805 h 2.5474285 2.5474283 v 2.41805 2.418062 H 9.6494677 7.1020392 Z m 5.9439998,0 v -2.41805 h 2.547428 2.547429 v 2.41805 2.418062 h -2.547429 -2.547428 z m 5.944,0 v -2.41805 h 2.547428 2.547428 v 2.41805 2.418062 H 21.537467 18.990039 Z M 7.1020392,21.312688 v -2.418061 h 2.5474285 2.5474283 v 2.418061 2.418062 H 9.6494677 7.1020392 Z m 5.9439998,0 v -2.418061 h 2.547428 2.547429 v 2.418061 2.418062 h -2.547429 -2.547428 z m 0,-5.642131 v -2.418062 h 2.547428 2.547429 v 2.418062 2.418049 h -2.547429 -2.547428 z"
	SVGIcons["deliver"] = "M 6.4718389,29.765572 C 5.5987968,29.444295 4.9364481,28.766538 4.6854844,27.937661 4.5034731,27.336519 4.5654572,26.499805 4.8326558,25.951018 c 0.3974755,-0.816359 1.166866,-1.395507 2.1027481,-1.582816 0.3655071,-0.07315 0.4980368,-0.06864 0.9005065,0.03069 0.9778424,0.241306 1.5926578,0.718966 2.0010366,1.554631 0.222594,0.455496 0.238118,0.532975 0.238118,1.188456 0,0.651266 -0.01637,0.734438 -0.2298672,1.168052 -0.4877754,0.990668 -1.3138949,1.534422 -2.3926306,1.574834 -0.4657078,0.01745 -0.6800003,-0.0087 -0.9807283,-0.119286 z m 1.3824593,-1.680227 c 0.710145,-0.432982 0.651251,-1.46157 -0.1046371,-1.82749 -0.3456529,-0.167327 -0.5998358,-0.14646 -0.9649952,0.07922 -0.4189536,0.258929 -0.5903202,0.968686 -0.3373004,1.397015 0.2648617,0.448375 0.9621302,0.622455 1.4069327,0.351255 z m 15.8290658,1.680227 c -0.873042,-0.321277 -1.535391,-0.999034 -1.786354,-1.827911 -0.182012,-0.601142 -0.120028,-1.437856 0.14717,-1.986643 0.397476,-0.816359 1.166866,-1.395507 2.102748,-1.582816 0.365507,-0.07315 0.498037,-0.06864 0.900507,0.03069 0.977842,0.241306 1.592658,0.718966 2.001036,1.554631 0.222594,0.455496 0.238119,0.532975 0.238119,1.188456 0,0.651266 -0.01637,0.734438 -0.229868,1.168052 -0.487776,0.990668 -1.313895,1.534422 -2.39263,1.574834 -0.465708,0.01745 -0.680001,-0.0087 -0.980728,-0.119286 z m 1.382459,-1.680227 c 0.710144,-0.432982 0.651251,-1.46157 -0.104638,-1.82749 -0.345652,-0.167327 -0.599835,-0.14646 -0.964994,0.07922 -0.418954,0.258929 -0.590321,0.968686 -0.337302,1.397015 0.264863,0.448375 0.962131,0.622455 1.406934,0.351255 z M 20.534914,27.462644 20.36,27.413984 20.34208,21.94233 20.32416,16.470676 h 2.5857 2.5857 l 1.595167,1.130857 c 0.877343,0.621972 2.019236,1.432697 2.537539,1.801613 0.518301,0.368916 1.00631,0.703148 1.084465,0.74274 0.137122,0.06946 0.143322,0.149073 0.17708,2.273878 l 0.03498,2.201898 0.244599,0.02321 c 0.134529,0.01276 0.28408,0.05597 0.332336,0.09603 0.121814,0.1011 0.125274,2.50892 0.0038,2.630415 -0.06067,0.06068 -0.552903,0.08396 -1.774416,0.08396 h -1.690458 l -0.03625,-0.635743 c -0.03216,-0.564174 -0.0683,-0.700424 -0.320968,-1.21033 -0.671874,-1.355898 -2.127335,-2.137925 -3.592584,-1.930317 -1.117199,0.158295 -1.940339,0.712773 -2.601741,1.752571 -0.204403,0.321343 -0.464421,1.198842 -0.464421,1.567305 0,0.486524 -0.09269,0.57432 -0.489759,0.463881 z m 7.06652,-6.193896 V 20.399601 L 27.003875,20.026859 C 26.675218,19.821853 26.084884,19.458328 25.692021,19.21903 l -0.714295,-0.435089 -1.031992,-0.0022 -1.031992,-0.0022 v 1.679174 1.679172 h 2.343847 2.343845 z M 0.3942902,27.361719 c -0.0557924,-0.06723 -0.0784076,-2.226216 -0.0803534,-7.671067 -0.001489,-4.167631 0.0119795,-7.592197 0.029931,-7.610148 0.0430602,-0.04306 1.6487482,-0.0424 1.6918439,7e-4 0.018336,0.01834 0.032374,2.422401 0.031197,5.342368 -0.00117,2.919967 0.023246,5.348526 0.054274,5.396799 0.044791,0.06968 1.6343147,0.0841 7.7133233,0.06997 l 7.656906,-0.01781 0.01793,-5.404839 0.01793,-5.404838 0.156991,-0.0048 c 0.577012,-0.01756 1.607069,-0.0028 1.633307,0.02348 0.01737,0.01737 0.03037,3.441465 0.02887,7.609095 -0.0019,5.444852 -0.02456,7.603843 -0.08036,7.671068 -0.06323,0.07619 -0.85359,0.09356 -4.257308,0.09356 h -4.179661 l -0.03625,-0.635743 c -0.03217,-0.564173 -0.0683,-0.700423 -0.32097,-1.210329 -0.6718742,-1.355898 -2.1273352,-2.137926 -3.5925848,-1.930317 -1.1171986,0.158294 -1.940338,0.712773 -2.6017407,1.752571 -0.2191699,0.34456 -0.4644217,1.198085 -0.4644217,1.61628 v 0.407539 H 2.1425417 c -1.30922384,0 -1.68740225,-0.02024 -1.7482515,-0.09356 z m 2.4533332,-5.237816 c -0.1194499,-0.11945 -0.1194499,-3.560738 0,-3.680188 0.1197338,-0.119734 3.9102814,-0.119734 4.0300157,0 0.11945,0.11945 0.11945,3.560738 0,3.680188 -0.1197343,0.119734 -3.9102819,0.119734 -4.0300157,0 z m 4.8975886,0 c -0.11945,-0.11945 -0.11945,-3.560738 0,-3.680188 0.1197343,-0.119734 3.910281,-0.119734 4.030015,0 0.11945,0.11945 0.11945,3.560738 0,3.680188 -0.119734,0.119734 -3.9102807,0.119734 -4.030015,0 z m 4.897588,0 c -0.11945,-0.11945 -0.11945,-3.560738 0,-3.680188 0.119734,-0.119734 3.910281,-0.119734 4.030015,0 0.11945,0.11945 0.11945,3.560738 0,3.680188 -0.119734,0.119734 -3.910281,0.119734 -4.030015,0 z M 2.8476234,17.576142 c -0.1194499,-0.11945 -0.1194499,-3.560738 0,-3.680188 0.1197338,-0.119734 3.9102814,-0.119734 4.0300157,0 0.11945,0.11945 0.11945,3.560738 0,3.680188 -0.1197343,0.119734 -3.9102819,0.119734 -4.0300157,0 z m 4.8975886,0 c -0.11945,-0.11945 -0.11945,-3.560738 0,-3.680188 0.1197343,-0.119734 3.910281,-0.119734 4.030015,0 0.11945,0.11945 0.11945,3.560738 0,3.680188 -0.119734,0.119734 -3.9102807,0.119734 -4.030015,0 z m 4.897588,0 c -0.11945,-0.11945 -0.11945,-3.560738 0,-3.680188 0.119734,-0.119734 3.910281,-0.119734 4.030015,0 0.11945,0.11945 0.11945,3.560738 0,3.680188 -0.119734,0.119734 -3.910281,0.119734 -4.030015,0 z"
	SVGIcons["female"] = "m 15.993191,1.74425 c -4.444127,0.0574 -8.0226427,3.73034 -7.9634743,8.17363 0.051997,3.92147 2.9721333,7.01558 6.7003793,7.68476 v 4.41529 h -3.725137 a 1.1559165,1.1559165 0 0 0 -1.1547035,1.15471 v 0.40703 a 1.1559165,1.1559165 0 0 0 1.1547035,1.1547 h 3.725137 v 2.931 a 1.1559165,1.1559165 0 0 0 1.154703,1.1547 h 0.407022 a 1.1559165,1.1559165 0 0 0 1.154704,-1.1547 v -2.931 h 3.928647 a 1.1559165,1.1559165 0 0 0 1.15028,-1.27416 l -0.04203,-0.40701 a 1.1559165,1.1559165 0 0 0 -1.15028,-1.03527 h -3.886618 v -4.40424 c 3.825396,-0.67182 6.778022,-3.90262 6.724711,-7.90596 -0.05892,-4.44322 -3.733989,-8.02112 -8.178044,-7.96348 z m 0.03096,2.30941 c 3.195235,-0.0414 5.795321,2.49141 5.837668,5.68505 0.04253,3.19366 -2.489749,5.79197 -5.685035,5.83324 -3.195234,0.0414 -5.7953194,-2.49141 -5.8376689,-5.68503 -0.042528,-3.19369 2.4897499,-5.79199 5.6850359,-5.83326 z"
	SVGIcons["male"] = "m 21.664275,1.91813 c -0.255266,0.0118 -0.435636,0.0851 -0.635157,0.24478 l -6.672617,5.34925 c -0.407447,0.32673 -0.558448,0.87852 -0.374165,1.36726 l 0.06005,0.15941 c 0.445885,0.7275 1.24199,0.93717 1.882374,0.51736 l 2.060221,-1.64456 -2.362786,4.6817 c -0.865655,-0.3161 -1.763353,-0.57637 -2.734641,-0.58891 -4.640107,-0.0603 -8.4773009,3.67547 -8.5388206,8.31473 -0.061776,4.63937 3.6746026,8.47437 8.3147836,8.5342 4.640108,0.0603 8.477305,-3.67548 8.538824,-8.31472 0.0362,-2.71833 -1.315184,-5.05078 -3.305128,-6.61955 l 2.591441,-5.05586 v 2.41595 c 7.2e-4,0.66553 0.540075,1.20489 1.205645,1.2056 h 0.297946 c 0.665568,-6.5e-4 1.204945,-0.54007 1.205643,-1.2056 v -8.08839 c 3.4e-4,-0.41105 -0.250119,-0.7709 -0.595892,-0.99323 l 0.309494,0.27944 c -0.115346,-0.20103 -0.364658,-0.33218 -0.570484,-0.43884 -0.113028,-0.0342 -0.304442,-0.0947 -0.422668,-0.0947 l 0.415739,0.0969 c -0.158384,-0.0639 -0.499332,-0.1295 -0.669802,-0.12239 z m -8.809054,12.4975 c 3.336247,0.0426 5.980238,2.75605 5.935833,6.09059 -0.04421,3.33456 -2.759007,5.9792 -6.095199,5.93591 -3.336245,-0.0426 -5.9802358,-2.75603 -5.9358309,-6.09059 0.044215,-3.33453 2.7590049,-5.97918 6.0951969,-5.93591 z"
	SVGIcons["arrowleftalt"] = "M16,30.534c8.027,0,14.534-6.507,14.534-14.534c0-8.027-6.507-14.534-14.534-14.534C7.973,1.466,1.466,7.973,1.466,16C1.466,24.027,7.973,30.534,16,30.534zM18.335,6.276l3.536,3.538l-6.187,6.187l6.187,6.187l-3.536,3.537l-9.723-9.724L18.335,6.276z"

	SVGIcons["question"] = "M16,1.466C7.973,1.466,1.466,7.973,1.466,16c0,8.027,6.507,14.534,14.534,14.534c8.027,0,14.534-6.507,14.534-14.534C30.534,7.973,24.027,1.466,16,1.466z M17.328,24.371h-2.707v-2.596h2.707V24.371zM17.328,19.003v0.858h-2.707v-1.057c0-3.19,3.63-3.696,3.63-5.963c0-1.034-0.924-1.826-2.134-1.826c-1.254,0-2.354,0.924-2.354,0.924l-1.541-1.915c0,0,1.519-1.584,4.137-1.584c2.487,0,4.796,1.54,4.796,4.136C21.156,16.208,17.328,16.627,17.328,19.003z"
	SVGIcons["i"] = "M16,1.466C7.973,1.466,1.466,7.973,1.466,16c0,8.027,6.507,14.534,14.534,14.534c8.027,0,14.534-6.507,14.534-14.534C30.534,7.973,24.027,1.466,16,1.466z M14.757,8h2.42v2.574h-2.42V8z M18.762,23.622H16.1c-1.034,0-1.475-0.44-1.475-1.496v-6.865c0-0.33-0.176-0.484-0.484-0.484h-0.88V12.4h2.662c1.035,0,1.474,0.462,1.474,1.496v6.887c0,0.309,0.176,0.484,0.484,0.484h0.88V23.622z"
	SVGIcons["$"] = "M16,1.466C7.973,1.466,1.466,7.973,1.466,16c0,8.027,6.507,14.534,14.534,14.534c8.027,0,14.534-6.507,14.534-14.534C30.534,7.973,24.027,1.466,16,1.466z M17.255,23.88v2.047h-1.958v-2.024c-3.213-0.44-4.621-3.08-4.621-3.08l2.002-1.673c0,0,1.276,2.223,3.586,2.223c1.276,0,2.244-0.683,2.244-1.849c0-2.729-7.349-2.398-7.349-7.459c0-2.2,1.738-3.785,4.137-4.159V5.859h1.958v2.046c1.672,0.22,3.652,1.1,3.652,2.993v1.452h-2.596v-0.704c0-0.726-0.925-1.21-1.959-1.21c-1.32,0-2.288,0.66-2.288,1.584c0,2.794,7.349,2.112,7.349,7.415C21.413,21.614,19.785,23.506,17.255,23.88z"
	SVGIcons["!"] = "M26.711,14.086L16.914,4.29c-0.778-0.778-2.051-0.778-2.829,0L4.29,14.086c-0.778,0.778-0.778,2.05,0,2.829l9.796,9.796c0.778,0.777,2.051,0.777,2.829,0l9.797-9.797C27.488,16.136,27.488,14.864,26.711,14.086zM14.702,8.981c0.22-0.238,0.501-0.357,0.844-0.357s0.624,0.118,0.844,0.353c0.221,0.235,0.33,0.531,0.33,0.885c0,0.306-0.101,1.333-0.303,3.082c-0.201,1.749-0.379,3.439-0.531,5.072H15.17c-0.135-1.633-0.301-3.323-0.5-5.072c-0.198-1.749-0.298-2.776-0.298-3.082C14.372,9.513,14.482,9.22,14.702,8.981zM16.431,21.799c-0.247,0.241-0.542,0.362-0.885,0.362s-0.638-0.121-0.885-0.362c-0.248-0.241-0.372-0.533-0.372-0.876s0.124-0.638,0.372-0.885c0.247-0.248,0.542-0.372,0.885-0.372s0.638,0.124,0.885,0.372c0.248,0.247,0.372,0.542,0.372,0.885S16.679,21.558,16.431,21.799z"
	SVGIcons["arrowalt"] = "M16,1.466C7.973,1.466,1.466,7.973,1.466,16c0,8.027,6.507,14.534,14.534,14.534c8.027,0,14.534-6.507,14.534-14.534C30.534,7.973,24.027,1.466,16,1.466zM13.665,25.725l-3.536-3.539l6.187-6.187l-6.187-6.187l3.536-3.536l9.724,9.723L13.665,25.725z"
	SVGIcons["question2"] = "M26.711,14.086L16.914,4.29c-0.778-0.778-2.051-0.778-2.829,0L4.29,14.086c-0.778,0.778-0.778,2.05,0,2.829l9.796,9.796c0.778,0.777,2.051,0.777,2.829,0l9.797-9.797C27.488,16.136,27.488,14.864,26.711,14.086zM16.431,21.799c-0.248,0.241-0.543,0.362-0.885,0.362c-0.343,0-0.638-0.121-0.886-0.362c-0.247-0.241-0.371-0.533-0.371-0.876s0.124-0.638,0.371-0.885c0.248-0.248,0.543-0.372,0.886-0.372c0.342,0,0.637,0.124,0.885,0.372c0.248,0.247,0.371,0.542,0.371,0.885S16.679,21.558,16.431,21.799zM18.911,15.198c-0.721,0.716-1.712,1.147-2.972,1.294v2.027h-0.844v-3.476c0.386-0.03,0.768-0.093,1.146-0.188c0.38-0.095,0.719-0.25,1.019-0.464c0.312-0.227,0.555-0.5,0.729-0.822c0.174-0.322,0.261-0.77,0.261-1.346c0-0.918-0.194-1.623-0.582-2.113c-0.389-0.49-0.956-0.735-1.701-0.735c-0.281,0-0.527,0.042-0.738,0.124s-0.366,0.16-0.464,0.234c0.031,0.146,0.072,0.357,0.124,0.633c0.052,0.275,0.078,0.486,0.078,0.633c0,0.226-0.098,0.433-0.294,0.619c-0.195,0.187-0.479,0.28-0.853,0.28c-0.33,0-0.565-0.113-0.706-0.339s-0.211-0.489-0.211-0.789c0-0.244,0.067-0.484,0.201-0.72c0.135-0.235,0.346-0.463,0.633-0.684c0.245-0.195,0.577-0.364,0.995-0.504c0.419-0.141,0.854-0.211,1.308-0.211c0.647,0,1.223,0.103,1.724,0.308c0.502,0.205,0.914,0.479,1.238,0.822c0.337,0.355,0.586,0.755,0.748,1.198c0.162,0.444,0.243,0.926,0.243,1.446C19.994,13.558,19.633,14.482,18.911,15.198z"
	SVGIcons["stopsign"] = "M20.833,2.625H10.167l-7.542,7.542v10.666l7.542,7.542h10.666l7.542-7.542V10.167L20.833,2.625zM23.76,17.145c-0.461,0.432-0.496-0.021-1.151,0.385c-0.655,0.404-1.067,0.844-1.067,0.844l-0.578,0.564c0,0-1.199,1.422-1.59,1.717c-0.392,0.291-0.572,0.166-0.572,0.166c-0.133,0.23-0.641,0.656-1.193,1.025c-0.549,0.371-0.99,1.354-0.99,1.354l-0.208,1.465c0,0-1.597,0.021-3.096-0.293c-1.499-0.314-2.699-0.982-2.699-0.982l0.655-1.451c0,0-0.58-2.225-0.636-3.926c-0.056-1.703,0.572-2.958,0.481-3.37c-0.09-0.412-0.438-1.109-0.523-1.569c-0.083-0.46-0.307-2.036-0.327-2.238c-0.021-0.202-0.28-2.232,0.614-2.204c0.891,0.027,0.648,1.388,0.725,2.246c0.077,0.857,1.13,3.249,1.297,3.123c0.167-0.125-0.056-2.398,0.006-2.838c0.063-0.439,0.182-2.315,0.293-2.747c0.112-0.433,0.105-1.778,0.935-1.66c0.83,0.118,0.606,1.332,0.488,1.813c-0.118,0.48,0.021,1.597,0.07,2.302c0.048,0.704,0.112,2.356,0.112,2.356l0.404-0.042c0,0,0.913-5,0.976-5.439c0.062-0.439,0.182-1.618,0.858-1.472c1.117,0.242,0.516,1.967,0.516,1.967l-0.486,5.509c0,0,0.166,0.224,0.492,0.021c0.33-0.202,1.312-3.25,1.459-3.927c0.146-0.676,0.084-2.678,1.145-2.428c1.059,0.252,0.453,2.755,0.124,4.353c-0.327,1.597-1.03,3.39-1.03,3.39l-0.433,1.631l0.376,0.852l0.412-0.092l0.898-0.717c0,0,0,0,0.614-0.461c0.613-0.46,1.764-0.495,2.044-0.355C23.453,16.154,24.221,16.713,23.76,17.145z"
	SVGIcons["temp"] = "M17.5,19.508V8.626h-3.999v10.881c-1.404,0.727-2.375,2.178-2.375,3.869c0,2.416,1.959,4.375,4.375,4.375s4.375-1.959,4.375-4.375C19.876,21.686,18.905,20.234,17.5,19.508zM20.5,5.249c0-2.757-2.244-5-5.001-5s-4.998,2.244-4.998,5v12.726c-1.497,1.373-2.376,3.314-2.376,5.4c0,4.066,3.31,7.377,7.376,7.377s7.374-3.311,7.374-7.377c0-2.086-0.878-4.029-2.375-5.402V5.249zM20.875,23.377c0,2.963-2.41,5.373-5.375,5.373c-2.962,0-5.373-2.41-5.373-5.373c0-1.795,0.896-3.443,2.376-4.438V5.251c0-1.654,1.343-3,2.997-3s3,1.345,3,3v13.688C19.979,19.934,20.875,21.582,20.875,23.377zM22.084,8.626l4.5,2.598V6.029L22.084,8.626z"
	SVGIcons["thunder"] = "M25.371,7.306c-0.092-3.924-3.301-7.077-7.248-7.079c-2.638,0.001-4.942,1.412-6.208,3.517c-0.595-0.327-1.28-0.517-2.01-0.517C7.626,3.229,5.772,5.033,5.689,7.293c-2.393,0.786-4.125,3.025-4.127,5.686c0,3.312,2.687,6,6,6v-0.002h5.271l-2.166,3.398l1.977-0.411L10,30.875l9.138-10.102L17,21l2.167-2.023h4.269c3.312,0,6-2.688,6-6C29.434,10.34,27.732,8.11,25.371,7.306zM23.436,16.979H7.561c-2.209-0.006-3.997-1.792-4.001-4.001c-0.002-1.982,1.45-3.618,3.35-3.931c0.265-0.043,0.502-0.191,0.657-0.414C7.722,8.41,7.779,8.136,7.73,7.87C7.702,7.722,7.685,7.582,7.685,7.446C7.689,6.221,8.68,5.23,9.905,5.228c0.647,0,1.217,0.278,1.633,0.731c0.233,0.257,0.587,0.375,0.927,0.309c0.342-0.066,0.626-0.307,0.748-0.63c0.749-1.992,2.662-3.412,4.911-3.41c2.899,0.004,5.244,2.35,5.251,5.249c0,0.161-0.009,0.326-0.027,0.497c-0.049,0.517,0.305,0.984,0.815,1.079c1.86,0.344,3.274,1.966,3.271,3.923C27.43,15.186,25.645,16.973,23.436,16.979z"
	SVGIcons["snow"] = "M25.372,6.912c-0.093-3.925-3.302-7.078-7.248-7.08c-2.638,0.002-4.942,1.412-6.208,3.518c-0.595-0.327-1.28-0.518-2.01-0.518C7.627,2.834,5.773,4.639,5.69,6.898c-2.393,0.786-4.125,3.025-4.127,5.686c0,3.312,2.687,6,6,6v-0.002h15.875c3.312,0,6-2.688,6-6C29.434,9.944,27.732,7.715,25.372,6.912zM23.436,16.584H7.562c-2.209-0.006-3.997-1.793-4.001-4c-0.002-1.983,1.45-3.619,3.35-3.933c0.265-0.043,0.502-0.19,0.657-0.414C7.723,8.015,7.78,7.74,7.731,7.475C7.703,7.326,7.686,7.187,7.686,7.051c0.004-1.225,0.995-2.217,2.22-2.219c0.647,0,1.217,0.278,1.633,0.731c0.233,0.257,0.587,0.375,0.927,0.31c0.342-0.066,0.626-0.308,0.748-0.631c0.749-1.992,2.662-3.412,4.911-3.41c2.898,0.004,5.244,2.351,5.251,5.25c0,0.16-0.009,0.325-0.026,0.496c-0.05,0.518,0.305,0.984,0.814,1.079c1.859,0.345,3.273,1.966,3.271,3.923C27.43,14.791,25.645,16.578,23.436,16.584zM16.667,24.09l1.119-1.119c0.389-0.391,0.389-1.025,0-1.416c-0.392-0.391-1.025-0.391-1.415,0l-1.119,1.119l-1.119-1.119c-0.391-0.391-1.025-0.391-1.415,0c-0.391,0.391-0.391,1.025,0,1.416l1.118,1.117l-1.12,1.121c-0.389,0.393-0.389,1.021,0,1.414c0.195,0.188,0.451,0.293,0.707,0.293c0.256,0,0.512-0.104,0.708-0.293l1.12-1.119l1.12,1.119c0.195,0.188,0.451,0.293,0.708,0.293c0.256,0,0.512-0.104,0.707-0.293c0.391-0.396,0.391-1.021,0-1.414L16.667,24.09zM25.119,21.817c-0.393-0.392-1.025-0.392-1.415,0l-1.12,1.121l-1.12-1.121c-0.391-0.392-1.022-0.392-1.414,0c-0.39,0.392-0.39,1.022,0,1.416l1.119,1.119l-1.119,1.119c-0.39,0.391-0.39,1.022,0,1.413c0.195,0.195,0.451,0.294,0.707,0.294c0.257,0,0.513-0.099,0.707-0.294l1.12-1.118l1.12,1.118c0.194,0.195,0.45,0.294,0.707,0.294c0.256,0,0.513-0.099,0.708-0.294c0.389-0.391,0.389-1.022,0-1.413l-1.12-1.119l1.12-1.119C25.507,22.842,25.507,22.209,25.119,21.817zM9.334,23.953l1.119-1.119c0.389-0.394,0.389-1.021,0-1.414c-0.391-0.394-1.025-0.394-1.415,0l-1.119,1.119l-1.12-1.121c-0.391-0.39-1.023-0.39-1.415,0c-0.391,0.396-0.391,1.024,0,1.418l1.119,1.117l-1.12,1.118c-0.391,0.394-0.391,1.025,0,1.414c0.196,0.195,0.452,0.293,0.708,0.293c0.256,0,0.511-0.098,0.707-0.293l1.12-1.119l1.121,1.121c0.195,0.195,0.451,0.293,0.707,0.293s0.513-0.098,0.708-0.293c0.389-0.391,0.389-1.022,0-1.416L9.334,23.953z"
	SVGIcons["hail"] = "M25.372,6.912c-0.093-3.925-3.302-7.078-7.248-7.08c-2.638,0.002-4.942,1.412-6.208,3.518c-0.595-0.327-1.28-0.518-2.01-0.518C7.627,2.834,5.773,4.639,5.69,6.898c-2.393,0.786-4.125,3.025-4.127,5.686c0,3.312,2.687,6,6,6v-0.002h15.875c3.312,0,6-2.688,6-6C29.434,9.944,27.732,7.715,25.372,6.912zM23.436,16.584H7.562c-2.209-0.006-3.997-1.793-4.001-4c-0.002-1.983,1.45-3.619,3.35-3.933c0.265-0.043,0.502-0.19,0.657-0.414C7.723,8.015,7.78,7.74,7.731,7.475C7.703,7.326,7.686,7.187,7.686,7.051c0.004-1.225,0.995-2.217,2.22-2.219c0.647,0,1.217,0.278,1.633,0.731c0.233,0.257,0.587,0.375,0.927,0.31c0.342-0.066,0.626-0.308,0.748-0.631c0.749-1.992,2.662-3.412,4.911-3.41c2.898,0.004,5.244,2.351,5.251,5.25c0,0.16-0.009,0.325-0.026,0.496c-0.05,0.518,0.305,0.984,0.814,1.079c1.859,0.345,3.273,1.966,3.271,3.923C27.43,14.791,25.645,16.578,23.436,16.584zM11.503,23.709c-0.784-0.002-1.418-0.636-1.418-1.416c0-0.785,0.634-1.416,1.418-1.418c0.78,0.002,1.413,0.633,1.416,1.418C12.917,23.073,12.284,23.707,11.503,23.709zM19.002,23.709c-0.783-0.002-1.418-0.636-1.418-1.416c0-0.785,0.635-1.416,1.418-1.418c0.779,0.002,1.414,0.633,1.414,1.418C20.417,23.073,19.784,23.707,19.002,23.709zM7.503,28.771c-0.783-0.002-1.417-0.637-1.417-1.418s0.634-1.414,1.417-1.416c0.78,0.002,1.415,0.635,1.415,1.416C8.917,28.135,8.284,28.77,7.503,28.771zM15.001,28.771c-0.782-0.002-1.417-0.637-1.417-1.418s0.634-1.414,1.417-1.416c0.78,0.002,1.413,0.635,1.415,1.416C16.415,28.135,15.784,28.77,15.001,28.771zM22.5,28.771c-0.782-0.002-1.416-0.634-1.416-1.416c0-0.785,0.634-1.418,1.416-1.42c0.781,0.002,1.414,0.635,1.418,1.42C23.915,28.138,23.282,28.77,22.5,28.771z"
	SVGIcons["rain"] = "M25.371,7.306c-0.092-3.924-3.301-7.077-7.248-7.079c-2.638,0.001-4.942,1.412-6.208,3.517c-0.595-0.327-1.28-0.517-2.01-0.517C7.626,3.229,5.772,5.033,5.689,7.293c-2.393,0.786-4.125,3.025-4.127,5.686c0,3.312,2.687,6,6,6v-0.002h15.874c3.312,0,6-2.688,6-6C29.434,10.34,27.732,8.11,25.371,7.306zM23.436,16.979H7.561c-2.209-0.006-3.997-1.792-4.001-4.001c-0.002-1.982,1.45-3.618,3.35-3.931c0.265-0.043,0.502-0.191,0.657-0.414C7.722,8.41,7.779,8.136,7.73,7.87C7.702,7.722,7.685,7.582,7.685,7.446C7.689,6.221,8.68,5.23,9.905,5.228c0.647,0,1.217,0.278,1.633,0.731c0.233,0.257,0.587,0.375,0.927,0.309c0.342-0.066,0.626-0.307,0.748-0.63c0.749-1.992,2.662-3.412,4.911-3.41c2.899,0.004,5.244,2.35,5.251,5.249c0,0.161-0.009,0.326-0.027,0.497c-0.049,0.517,0.305,0.984,0.815,1.079c1.86,0.344,3.274,1.966,3.271,3.923C27.43,15.186,25.645,16.973,23.436,16.979zM9.029,26.682c0-1.115,0.021-5.425,0.021-5.432c0.002-0.409-0.247-0.779-0.628-0.932c-0.38-0.152-0.815-0.059-1.099,0.24c-0.006,0.008-1.037,1.098-2.081,2.342c-0.523,0.627-1.048,1.287-1.463,1.896c-0.399,0.648-0.753,1.066-0.811,1.885C2.971,28.355,4.324,29.711,6,29.714C7.672,29.71,9.029,28.354,9.029,26.682zM4.971,26.727c0.091-0.349,1.081-1.719,1.993-2.764c0.025-0.029,0.051-0.061,0.076-0.089c-0.005,1.124-0.01,2.294-0.01,2.808c0,0.567-0.461,1.028-1.029,1.03C5.447,27.71,4.997,27.273,4.971,26.727zM16.425,26.682c0-1.115,0.021-5.424,0.021-5.43c0.002-0.41-0.247-0.779-0.628-0.934c-0.381-0.152-0.814-0.058-1.1,0.242c-0.006,0.008-1.035,1.094-2.08,2.342c-0.522,0.623-1.047,1.285-1.463,1.894c-0.399,0.649-0.753,1.068-0.809,1.888c0,1.672,1.354,3.028,3.029,3.028C15.068,29.711,16.425,28.354,16.425,26.682zM12.365,26.729c0.092-0.349,1.081-1.72,1.993-2.765c0.025-0.03,0.05-0.06,0.075-0.089c-0.005,1.123-0.011,2.294-0.011,2.807c-0.002,0.568-0.461,1.027-1.028,1.029C12.84,27.709,12.392,27.273,12.365,26.729zM23.271,20.317c-0.38-0.153-0.816-0.06-1.099,0.24c-0.009,0.008-1.037,1.097-2.08,2.342c-0.523,0.625-1.049,1.285-1.462,1.896c-0.402,0.649-0.754,1.067-0.812,1.886c0,1.672,1.354,3.029,3.03,3.029c1.673,0,3.027-1.357,3.027-3.029c0-1.115,0.022-5.425,0.022-5.431C23.9,20.84,23.651,20.47,23.271,20.317zM21.879,26.681c-0.004,0.568-0.463,1.027-1.031,1.029c-0.553-0.002-1.002-0.438-1.028-0.982c0.092-0.349,1.081-1.72,1.993-2.765c0.025-0.028,0.05-0.059,0.074-0.088C21.883,24.998,21.879,26.167,21.879,26.681z"
	SVGIcons["cloudy"] = "M14.378,6.781c0.41,0.988,1.938,0.346,1.524-0.648C15.708,5.667,15.515,5.2,15.32,4.734c-0.289-0.695-0.875-3.233-2.042-2.747c-1.03,0.433-0.128,1.846,0.142,2.494C13.739,5.248,14.059,6.015,14.378,6.781M20.8,7.223c1.094,0.453,1.538-1.551,1.813-2.216c0.281-0.677,1.478-2.565,0.357-3.029c-1.092-0.453-1.537,1.548-1.813,2.216C20.876,4.872,19.68,6.757,20.8,7.223M18.137,6.692c1.183,0,0.829-2.019,0.829-2.742c0-0.732,0.383-2.935-0.829-2.935c-1.183,0-0.828,2.019-0.828,2.742C17.309,4.49,16.926,6.692,18.137,6.692M23.058,8.729c0.852,0.85,2.142-0.972,2.659-1.49c0.512-0.513,2.187-1.687,1.352-2.524c-0.834-0.836-2.013,0.843-2.522,1.353C24.028,6.585,22.198,7.874,23.058,8.729M24.565,10.986c0.448,1.091,2.183-0.01,2.849-0.286c0.676-0.28,2.858-0.771,2.394-1.89c-0.455-1.091-2.181,0.008-2.849,0.285C26.281,9.377,24.102,9.866,24.565,10.986M12.036,8.742c0.752,0.75,1.932-0.415,1.17-1.173c-0.253-0.347-0.646-0.645-0.949-0.946c-0.541-0.539-2.162-2.799-3.068-1.889c-0.79,0.791,0.586,1.755,1.083,2.25C10.859,7.57,11.447,8.156,12.036,8.742M29.365,17.397c-0.768-0.317-1.534-0.635-2.302-0.952c-0.646-0.268-2.07-1.169-2.495-0.135c-0.481,1.168,2.054,1.747,2.751,2.035c0.455,0.188,0.911,0.377,1.367,0.565C29.7,19.331,30.379,17.816,29.365,17.397M29.942,12.817c-0.83,0-1.66,0-2.49,0c-0.701,0-2.357-0.288-2.355,0.83c0,1.262,2.567,0.827,3.319,0.827c0.493,0,0.986,0,1.479-0.001C30.99,14.473,31.043,12.815,29.942,12.817M24.234,18.568c-0.673-0.673-1.773,0.189-1.281,1.007c-0.295-0.264-0.614-0.499-0.961-0.69c3.894-2.866,3.328-9.006-1.021-11.107c-2.024-0.978-4.481-0.828-6.368,0.394c-0.871,0.564-1.603,1.336-2.119,2.236c-0.262,0.456-0.468,0.943-0.612,1.449c-0.074,0.258-0.131,0.521-0.172,0.786c-0.083,0.534-0.109,0.553-0.553,0.871c-0.182-0.957-1.64-0.675-2.326-0.674c-0.815,0.001-1.963-0.217-2.752,0.046c-0.867,0.289-0.652,1.615,0.263,1.613c0.324,0.052,0.701-0.001,1.028-0.001c0.904-0.001,1.809-0.002,2.713-0.003c-0.308,0.352-0.496,0.969-0.94,0.77c-0.467-0.209-0.978-0.319-1.49-0.319c-0.951,0-1.877,0.375-2.561,1.036c-0.681,0.658-1.088,1.569-1.123,2.516c-0.944,0.31-1.791,0.891-2.421,1.658c-2.756,3.354-0.265,8.554,4.058,8.554v-0.002c3.597,0,7.194,0,10.792,0c1.341,0,2.843,0.167,4.168-0.113c3.652-0.772,5.361-5.21,3.133-8.229c0.548,0.547,1.096,1.094,1.644,1.641c0.183,0.183,0.364,0.424,0.575,0.574c0.552,0.552,1.524,0.066,1.403-0.713c-0.097-0.622-1.042-1.267-1.448-1.673C25.319,19.652,24.776,19.11,24.234,18.568M18.137,8.787c4.559,0.009,6.576,5.979,2.912,8.734c-0.637-3.505-4.161-5.824-7.629-5.03C13.943,10.367,15.852,8.792,18.137,8.787M22.895,24.08c-0.633,3.346-4.149,2.879-6.68,2.879c-3.017,0-6.033,0-9.049,0c-0.767,0-1.62,0.084-2.373-0.095c-2.274-0.538-3.416-3.242-2.172-5.235c0.678-1.087,1.568-1.19,2.626-1.67c0.604-0.273,0.456-0.807,0.456-1.331c0.002-0.597,0.284-1.169,0.756-1.533c0.787-0.608,1.943-0.497,2.611,0.234c1.098,1.205,1.96-1.346,2.507-1.893c2.025-2.025,5.475-1.708,7.068,0.684c0.344,0.516,0.581,1.102,0.693,1.712c0.097,0.529-0.115,1.341,0.188,1.796c0.291,0.47,0.943,0.463,1.397,0.68c0.508,0.23,0.963,0.591,1.304,1.034C22.834,22.125,23.064,23.107,22.895,24.08M6.906,9.917c0.881,0.364,1.763,0.727,2.644,1.091c0.353,0.146,0.707,0.292,1.06,0.437c0.997,0.412,1.637-1.119,0.642-1.526C10.47,9.441,9.456,9.177,8.609,8.828c-0.354-0.146-0.707-0.292-1.06-0.437C6.554,7.98,5.912,9.505,6.906,9.917"
	SVGIcons["sun"] = "M15.502,7.504c-4.35,0-7.873,3.523-7.873,7.873c0,4.347,3.523,7.872,7.873,7.872c4.346,0,7.871-3.525,7.871-7.872C23.374,11.027,19.85,7.504,15.502,7.504zM15.502,21.25c-3.244-0.008-5.866-2.63-5.874-5.872c0.007-3.243,2.63-5.866,5.874-5.874c3.242,0.008,5.864,2.631,5.871,5.874C21.366,18.62,18.744,21.242,15.502,21.25zM15.502,6.977c0.553,0,1-0.448,1-1.001V1.125c-0.002-0.553-0.448-1-1-1c-0.553,0-1.001,0.449-1,1.002v4.85C14.502,6.528,14.949,6.977,15.502,6.977zM18.715,7.615c0.125,0.053,0.255,0.076,0.382,0.077c0.394,0,0.765-0.233,0.925-0.618l1.856-4.483c0.21-0.511-0.031-1.095-0.541-1.306c-0.511-0.211-1.096,0.031-1.308,0.541L18.174,6.31C17.963,6.82,18.205,7.405,18.715,7.615zM21.44,9.436c0.195,0.194,0.451,0.293,0.707,0.293s0.512-0.098,0.707-0.293l3.43-3.433c0.391-0.39,0.39-1.023,0-1.415c-0.392-0.39-1.025-0.39-1.415,0.002L21.44,8.021C21.049,8.412,21.049,9.045,21.44,9.436zM23.263,12.16c0.158,0.385,0.531,0.617,0.923,0.617c0.127,0,0.257-0.025,0.383-0.078l4.48-1.857c0.511-0.211,0.753-0.797,0.541-1.307s-0.796-0.752-1.307-0.54l-4.481,1.857C23.292,11.064,23.051,11.65,23.263,12.16zM29.752,14.371l-4.851,0.001c-0.552,0-1,0.448-0.998,1.001c0,0.553,0.447,0.999,0.998,0.999l4.852-0.002c0.553,0,0.999-0.449,0.999-1C30.752,14.817,30.304,14.369,29.752,14.371zM29.054,19.899l-4.482-1.854c-0.512-0.212-1.097,0.03-1.307,0.541c-0.211,0.511,0.031,1.096,0.541,1.308l4.482,1.854c0.126,0.051,0.256,0.075,0.383,0.075c0.393,0,0.765-0.232,0.925-0.617C29.806,20.695,29.563,20.109,29.054,19.899zM22.86,21.312c-0.391-0.391-1.023-0.391-1.414,0.001c-0.391,0.39-0.39,1.022,0,1.413l3.434,3.429c0.195,0.195,0.45,0.293,0.706,0.293s0.513-0.098,0.708-0.293c0.391-0.392,0.389-1.025,0-1.415L22.86,21.312zM20.029,23.675c-0.211-0.511-0.796-0.752-1.307-0.541c-0.51,0.212-0.752,0.797-0.54,1.308l1.86,4.48c0.159,0.385,0.531,0.617,0.925,0.617c0.128,0,0.258-0.024,0.383-0.076c0.511-0.211,0.752-0.797,0.54-1.309L20.029,23.675zM15.512,23.778c-0.553,0-1,0.448-1,1l0.004,4.851c0,0.553,0.449,0.999,1,0.999c0.553,0,1-0.448,0.998-1l-0.003-4.852C16.511,24.226,16.062,23.777,15.512,23.778zM12.296,23.142c-0.51-0.21-1.094,0.031-1.306,0.543l-1.852,4.483c-0.21,0.511,0.033,1.096,0.543,1.307c0.125,0.052,0.254,0.076,0.382,0.076c0.392,0,0.765-0.234,0.924-0.619l1.853-4.485C13.051,23.937,12.807,23.353,12.296,23.142zM9.57,21.325c-0.392-0.391-1.025-0.389-1.415,0.002L4.729,24.76c-0.391,0.392-0.389,1.023,0.002,1.415c0.195,0.194,0.45,0.292,0.706,0.292c0.257,0,0.513-0.098,0.708-0.293l3.427-3.434C9.961,22.349,9.961,21.716,9.57,21.325zM7.746,18.604c-0.213-0.509-0.797-0.751-1.307-0.54L1.96,19.925c-0.511,0.212-0.752,0.798-0.54,1.308c0.16,0.385,0.531,0.616,0.924,0.616c0.127,0,0.258-0.024,0.383-0.076l4.479-1.861C7.715,19.698,7.957,19.113,7.746,18.604zM7.1,15.392c0-0.553-0.447-0.999-1-0.999l-4.851,0.006c-0.553,0-1.001,0.448-0.999,1.001c0.001,0.551,0.449,1,1,0.998l4.852-0.006C6.654,16.392,7.102,15.942,7.1,15.392zM1.944,10.869l4.485,1.85c0.125,0.053,0.254,0.076,0.381,0.076c0.393,0,0.766-0.232,0.925-0.618c0.212-0.511-0.032-1.097-0.544-1.306L2.708,9.021c-0.511-0.21-1.095,0.032-1.306,0.542C1.19,10.074,1.435,10.657,1.944,10.869zM8.137,9.451c0.195,0.193,0.449,0.291,0.705,0.291s0.513-0.098,0.709-0.295c0.391-0.389,0.389-1.023-0.004-1.414L6.113,4.609C5.723,4.219,5.088,4.221,4.699,4.612c-0.391,0.39-0.389,1.024,0.002,1.414L8.137,9.451zM10.964,7.084c0.16,0.384,0.532,0.615,0.923,0.615c0.128,0,0.258-0.025,0.384-0.077c0.51-0.212,0.753-0.798,0.54-1.307l-1.864-4.479c-0.212-0.51-0.798-0.751-1.308-0.539C9.129,1.51,8.888,2.096,9.1,2.605L10.964,7.084z"
	SVGIcons["undo"] = "M12.981,9.073V6.817l-12.106,6.99l12.106,6.99v-2.422c3.285-0.002,9.052,0.28,9.052,2.269c0,2.78-6.023,4.263-6.023,4.263v2.132c0,0,13.53,0.463,13.53-9.823C29.54,9.134,17.952,8.831,12.981,9.073z"
	SVGIcons["detour"] = "M29.342,15.5l-7.556-4.363v2.614H18.75c-1.441-0.004-2.423,1.002-2.875,1.784c-0.735,1.222-1.056,2.561-1.441,3.522c-0.135,0.361-0.278,0.655-0.376,0.817c-1.626,0-0.998,0-2.768,0c-0.213-0.398-0.571-1.557-0.923-2.692c-0.237-0.676-0.5-1.381-1.013-2.071C8.878,14.43,7.89,13.726,6.75,13.75H2.812v3.499c0,0,0.358,0,1.031,0h2.741c0.008,0.013,0.018,0.028,0.029,0.046c0.291,0.401,0.634,1.663,1.031,2.888c0.218,0.623,0.455,1.262,0.92,1.897c0.417,0.614,1.319,1.293,2.383,1.293H11c2.25,0,1.249,0,3.374,0c0.696,0.01,1.371-0.286,1.809-0.657c1.439-1.338,1.608-2.886,2.13-4.127c0.218-0.608,0.453-1.115,0.605-1.314c0.006-0.01,0.012-0.018,0.018-0.025h2.85v2.614L29.342,15.5zM10.173,14.539c0.568,0.76,0.874,1.559,1.137,2.311c0.04,0.128,0.082,0.264,0.125,0.399h2.58c0.246-0.697,0.553-1.479,1.005-2.228c0.252-0.438,0.621-0.887,1.08-1.272H9.43C9.735,14.003,9.99,14.277,10.173,14.539z"
	SVGIcons["merge"] = "M29.342,15.5l-7.556-4.363v2.613h-1.411c-0.788-0.01-1.331-0.241-2.019-0.743c-1.021-0.745-2.094-2.181-3.551-3.568C13.367,8.06,11.291,6.73,8.5,6.749H2.812v3.5H8.5c2.231,0.012,3.441,1.185,5.07,2.934c0.697,0.753,1.428,1.58,2.324,2.323c-1.396,1.165-2.412,2.516-3.484,3.501c-1.183,1.081-2.202,1.723-3.912,1.741H2.813v3.5h5.716c3.752,0.001,6.035-2.319,7.619-4.066c0.817-0.895,1.537-1.691,2.209-2.191c0.686-0.502,1.23-0.732,2.017-0.742h1.412v2.614L29.342,15.5z"
	SVGIcons["split"] = "M21.786,20.698c-1.792-0.237-2.912-1.331-4.358-2.886c-0.697-0.751-1.428-1.577-2.324-2.319c1.396-1.165,2.411-2.519,3.483-3.503c1.01-0.92,1.901-1.519,3.199-1.688v2.574l7.556-4.363L21.786,4.15v2.652c-3.34,0.266-5.45,2.378-6.934,4.013c-0.819,0.896-1.537,1.692-2.212,2.192c-0.685,0.501-1.227,0.731-2.013,0.742c-0.001,0-0.002,0-0.003,0H2.812v3.5h0.001v0.001c0,0,0.046-0.001,0.136-0.001h7.677c0.786,0.011,1.33,0.241,2.017,0.743c1.021,0.743,2.095,2.181,3.552,3.568c1.312,1.258,3.162,2.46,5.592,2.649v2.664l7.556-4.36l-7.556-4.361V20.698z"
	SVGIcons["fork"] = "M13.741,10.249h8.045v2.627l7.556-4.363l-7.556-4.363v2.598H9.826C11.369,7.612,12.616,8.922,13.741,10.249zM21.786,20.654c-0.618-0.195-1.407-0.703-2.291-1.587c-1.79-1.756-3.712-4.675-5.731-7.227c-2.049-2.486-4.159-4.972-7.451-5.091h-3.5v3.5h3.5c0.656-0.027,1.683,0.486,2.879,1.683c1.788,1.753,3.712,4.674,5.731,7.226c1.921,2.331,3.907,4.639,6.863,5.016v2.702l7.556-4.362l-7.556-4.362V20.654z"
	SVGIcons["exchange"] = "M21.786,12.876l7.556-4.363l-7.556-4.363v2.598H2.813v3.5h18.973V12.876zM10.368,18.124l-7.556,4.362l7.556,4.362V24.25h18.974v-3.501H10.368V18.124z"
	SVGIcons["shuffle"] = "M21.786,20.654c-0.618-0.195-1.407-0.703-2.291-1.587c-0.757-0.742-1.539-1.698-2.34-2.741c-0.191,0.256-0.382,0.51-0.574,0.77c-0.524,0.709-1.059,1.424-1.604,2.127c1.904,2.31,3.88,4.578,6.809,4.952v2.701l7.556-4.362l-7.556-4.362V20.654zM9.192,11.933c0.756,0.741,1.538,1.697,2.339,2.739c0.195-0.262,0.39-0.521,0.587-0.788c0.52-0.703,1.051-1.412,1.592-2.11c-2.032-2.463-4.133-4.907-7.396-5.025h-3.5v3.5h3.5C6.969,10.223,7.996,10.735,9.192,11.933zM21.786,10.341v2.535l7.556-4.363l-7.556-4.363v2.647c-1.904,0.219-3.425,1.348-4.751,2.644c-2.196,2.183-4.116,5.167-6.011,7.538c-1.867,2.438-3.741,3.888-4.712,3.771h-3.5v3.5h3.5c2.185-0.029,3.879-1.266,5.34-2.693c2.194-2.184,4.116-5.167,6.009-7.538C19.205,12.003,20.746,10.679,21.786,10.341z"
	SVGIcons["refresh"] = "M24.083,15.5c-0.009,4.739-3.844,8.574-8.583,8.583c-4.741-0.009-8.577-3.844-8.585-8.583c0.008-4.741,3.844-8.577,8.585-8.585c1.913,0,3.665,0.629,5.09,1.686l-1.782,1.783l8.429,2.256l-2.26-8.427l-1.89,1.89c-2.072-1.677-4.717-2.688-7.587-2.688C8.826,3.418,3.418,8.826,3.416,15.5C3.418,22.175,8.826,27.583,15.5,27.583S27.583,22.175,27.583,15.5H24.083z"
	SVGIcons["ccw"] = "M24.249,15.499c-0.009,4.832-3.918,8.741-8.75,8.75c-2.515,0-4.768-1.064-6.365-2.763l2.068-1.442l-7.901-3.703l0.744,8.694l2.193-1.529c2.244,2.594,5.562,4.242,9.26,4.242c6.767,0,12.249-5.482,12.249-12.249H24.249zM15.499,6.75c2.516,0,4.769,1.065,6.367,2.764l-2.068,1.443l7.901,3.701l-0.746-8.693l-2.192,1.529c-2.245-2.594-5.562-4.245-9.262-4.245C8.734,3.25,3.25,8.734,3.249,15.499H6.75C6.758,10.668,10.668,6.758,15.499,6.75z"
	SVGIcons["acw"] = "M19.275,3.849l1.695,8.56l1.875-1.642c2.311,3.59,1.72,8.415-1.584,11.317c-2.24,1.96-5.186,2.57-7.875,1.908l-0.84,3.396c3.75,0.931,7.891,0.066,11.02-2.672c4.768-4.173,5.521-11.219,1.94-16.279l2.028-1.775L19.275,3.849zM8.154,20.232c-2.312-3.589-1.721-8.416,1.582-11.317c2.239-1.959,5.186-2.572,7.875-1.909l0.842-3.398c-3.752-0.93-7.893-0.067-11.022,2.672c-4.765,4.174-5.519,11.223-1.939,16.283l-2.026,1.772l8.26,2.812l-1.693-8.559L8.154,20.232z"
	SVGIcons["contract"] = "M25.083,18.895l-8.428-2.259l2.258,8.428l1.838-1.837l7.053,7.053l2.476-2.476l-7.053-7.053L25.083,18.895zM5.542,11.731l8.428,2.258l-2.258-8.428L9.874,7.398L3.196,0.72L0.72,3.196l6.678,6.678L5.542,11.731zM7.589,20.935l-6.87,6.869l2.476,2.476l6.869-6.869l1.858,1.857l2.258-8.428l-8.428,2.258L7.589,20.935zM23.412,10.064l6.867-6.87l-2.476-2.476l-6.868,6.869l-1.856-1.856l-2.258,8.428l8.428-2.259L23.412,10.064z"
	//
	SVGIcons["home"] = "M27.812,16l-3.062-3.062V5.625h-2.625v4.688L16,4.188L4.188,16L7,15.933v11.942h17.875V16H27.812zM16,26.167h-5.833v-7H16V26.167zM21.667,23.167h-3.833v-4.042h3.833V23.167z"
	SVGIcons["user"] = "M20.771,12.364c0,0,0.849-3.51,0-4.699c-0.85-1.189-1.189-1.981-3.058-2.548s-1.188-0.454-2.547-0.396c-1.359,0.057-2.492,0.792-2.492,1.188c0,0-0.849,0.057-1.188,0.397c-0.34,0.34-0.906,1.924-0.906,2.321s0.283,3.058,0.566,3.624l-0.337,0.113c-0.283,3.283,1.132,3.68,1.132,3.68c0.509,3.058,1.019,1.756,1.019,2.548s-0.51,0.51-0.51,0.51s-0.452,1.245-1.584,1.698c-1.132,0.452-7.416,2.886-7.927,3.396c-0.511,0.511-0.453,2.888-0.453,2.888h26.947c0,0,0.059-2.377-0.452-2.888c-0.512-0.511-6.796-2.944-7.928-3.396c-1.132-0.453-1.584-1.698-1.584-1.698s-0.51,0.282-0.51-0.51s0.51,0.51,1.02-2.548c0,0,1.414-0.397,1.132-3.68H20.771z"
	//
	SVG_object.Total = "0"
}
